//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
texture2D pcaTexture;

//Samplers -------------------------------------------------------------------------

sampler2D pcatex = sampler_state
{
	Texture = <pcaTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};

//----------------------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 EigenPS(PixelShaderInput input) : COLOR0
{
	//calculation of loss of color contrast (in a least-squares sense) in the chromaticity plane (i.e., a*b* plane)

	float4 color = tex2D(pcatex, input.texC);	

	//shift from [0 1] to [-0.5 0.5]
	color.xyz = color.xyz -0.5f;
	//color.xyz = color.xyz *10000.0f-2000.0f; //correct??? needed?
	
	// compute EigenVector of matrix M = [a b; b c] , L* = only luminance -> not needed
	float a = color.x;
	float b = color.y;
	float c = color.z;

	
	//debug:
	/*if((color.x <= 0.20f) && (color.y <= 0.20f) && (color.z <= 0.20f))
	{
		color.xyz = (0.0f, 0.0f, 0.0f);
	}
	else if((color.x < 0.1f) || (color.y < 0.1f) || (color.z < 0.1f))
	{
		color.xyz = (0.0f, 0.0f, 0.0f);
	}
	return color;*/


	
	// calculate EigenValues of M using det(M) = 0
	/* | a-lamda     b    |
	   |    b     c-lamda |		= 0

	   => (a-lamda)*(c-lamda)-b*b = 0
	   solve this equation for lamda using the pq-rule: 
	   x^2 + px + q = 0 ...=>... 1x2= p/2 +- sqrt[(p/2)^2-q]	*/
	   
	float lamda1 = 0.5*(+sqrt( pow(a,2.0f) - 2*a*c + 4*pow(b,2.0f) + pow(c,2.0f) )+a+c); 
	float lamda2 = 0.5*(-sqrt( pow(a,2.0f) - 2*a*c + 4*pow(b,2.0f) + pow(c,2.0f) )+a+c); 

	// choose biggest EigenValue
	float lamda = 0.0f;
	if(abs(lamda1)>abs(lamda2)) { lamda=lamda1;	}
	else { lamda=lamda2; }

	// calculate EigenVector corresponding to biggest EigenValue
	//float4 outColor = float4(1.0f,1.0f,lamda,1.0f);	// output lamda only for debugging
	float4 outColor = float4(1.0f,1.0f,1.0f,1.0f);
	// In order to solve the resulting characteristic equation, we set the b* coordinate of 
	// vab to one and solve for its a* coordinate.
	/* ( a b ) (x)			(x)
	   ( b c ) (y) = lamda* (y)   => solve   a*x+b*y=lamda*x   for x with y=1	
								  or solve   b*x+c*y=lamda*y   for x with y=1*/
	// version1: outColor.x = b/(lamda-a);
	outColor.x = (lamda-c)/b;
	outColor.y = 1.0f;
	
	// scale to [-1 1]
	if(abs(outColor.x)>abs(outColor.y)) { outColor.xy=outColor.xy/outColor.x;	}
	else { outColor.xy=outColor.xy/outColor.y; }

	// shift and scale to [0 1] 
	outColor.xyz = outColor.xyz/2.0f +0.5f;

	if((outColor.x < 0.0f) || (outColor.y < 0.0f) || (outColor.z < 0.0f))
	{
		outColor.xyz = float3(0.0f, 0.0f, 1.0f);
	}

	if((outColor.x > 1.0f) || (outColor.y > 1.0f) || (outColor.z > 1.0f))
	{
		outColor.xyz = float3(0.0f, 1.0f, 0.0f);
	}

	//outColor.x = input.texC.x;
	outColor.w = 1.0f;
	return outColor;
}


technique EigenVector
{
    pass Pass1
    {		
        PixelShader = compile ps_2_0 EigenPS();
    }
}

