//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
texture2D projEffectSumTexture;
texture2D projEffectTexture; 
texture2D labTexture;
float2 viewport;

//Samplers -------------------------------------------------------------------------

sampler2D projsumtex = sampler_state
{
	Texture = <projEffectSumTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

sampler2D projtex = sampler_state
{
	Texture = <projEffectTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

sampler2D labtex = sampler_state
{
	Texture = <labTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};

//----------------------------------------------------------------------------------


////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader
////////////////////////////////////////////////////////////////////////////////////

void SpriteVertexShader(inout float4 color    : COLOR0,

                       inout float2 texCoord : TEXCOORD0,

                       inout float4 position : POSITION0)

{

   // Half pixel offset for correct texel centering.

   position.xy -= 0.5f;

   // Viewport adjustment.

   position.xy = position.xy / viewport;

   position.xy *= float2(2.0f, -2.0f);

   position.xy -= float2(1.0f, -1.0f);

}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 ResultPS(PixelShaderInput input) : COLOR0
{
	float2 o = float2(0.0f,0.0f);

	// ... and shift from [0 1] to actuel values
	float4 pes = tex2D( projsumtex, o)-0.5;
	float4 pe = tex2D( projtex, input.texC)-0.5;

	float4 color = tex2D(labtex, input.texC)-0.5; 
	// convert from [0,+1] to [0,510] and shift to [-255,255]:
	//color.xyz = (color.xyz*510.0f)-255.0f;

	// enhance contrast
	// pes.y = max dist vom origin of original
	// pes.z = max dist from origin of projected
	// -> scales all colors (b* coordinates) to max contrast 
	//color.z = pe.z ;
	color.z = pe.z /pes.z;
	//color.z = pe.z * pes.y/pes.z;
	color.z = color.z/2.0f; // sclae [-1 1] to [-0.5 0.5]

	color.y = 0.0f; // not needded?????
	//color.z = pes.y*(pe.x/pes.x);

	// shift [-0.5, 0.5] to [0 1]
	color.xyz = color.xyz+0.5; 

	
	/*if((color.x > 1.0f) || (color.y > 1.0f) || (color.z > 1.0f))
	{
		color.xyz = float3(0.0f, 0.0f, 1.0f);
	}*/

	//color.x = pe.z * pes.y/pes.z;;
	//color.z = pe.z*(10.0);

	/*if((color.z < 0.0f))
	{
		color.xyz = float3(0.0f, 1.0f, 0.0f);
	}
	else
	{
		color.xyz = float3(0.0f, 0.0f, 1.0f);
	}	*/

	color.w = 1.0f;
	return color;
}

technique Result
{
    pass Pass1
    {		
		VertexShader = compile vs_3_0 SpriteVertexShader();
        PixelShader = compile ps_3_0 ResultPS();
    }
}



