//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
texture2D labTexture;
float2 viewport;

//Samplers -------------------------------------------------------------------------

sampler2D labtex = sampler_state
{
	Texture = <labTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};


////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 LMSPS(PixelShaderInput input) : COLOR0
{
	// get color of pixel (in L*a*b* space)	
	float4 color = tex2D(labtex, input.texC); 
	
	// convert from [0,+1] to [-0.5,0.5] 
	color.xyz = (color.xyz)-0.5f;
	
	// project colors onto b* plane (a*=y=0)
	color.y = 0.0f;
	
	// shift [-0.5, 0.5] to [0 1]
	color.xyz = (color.xyz+0.5f); 

	//debug:
	if((color.x > 1.0f) || (color.y > 1.0f) || (color.z > 1.0f))
	{
		color.xyz = float3(0.0f, 0.0f, 1.0f);
	}
	else if((color.x < 0.0f) || (color.y < 0.0f) || (color.z < 0.0f))
	{
		color.xyz = float3(0.0f, 1.0f, 0.0f);
	}

	color.w = 1.0f;
	return color;
}


technique LMS
{
    pass Pass1
    {		
		PixelShader = compile ps_2_0 LMSPS();
    }
}

