//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
float4x4 rotmat; 
float2 viewport;
texture2D ResCtexture; //result

// XYZ to RGB matrix conversion
float4x4 mXYZtoRGB;

//Samplers -------------------------------------------------------------------------

sampler2D labtex = sampler_state
{
	Texture = <ResCtexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct VertexShaderOutput
{
    float2 texC			: TEXCOORD0;
};

//----------------------------------------------------------------------------------

float invF(float p)
{
	float p3 = pow(p,3.0f);
	if (p3 > 0.008856f)
	{
		return p3;
	}
	else
	{
		return (116.0f * p - 16.0f)/903.3f;
	}
}

// Convert input L*a*b* to CIE XYZ color space
float3 LabToXYZ(float3 vec3)
{
	float L = vec3.x;
	float a = vec3.y;
	float b = vec3.z; 

	float fy = (L+16.0f)/116.0f;
	float fx = (a/500.0f) + fy;
	float fz = fy - (b/200.0f);

	float xr = invF(fx); 
	float yr = invF(fy);
	float zr = invF(fz);

	// multiply by reference white
	float3 white = ( 0.9642f, 1.0f, 0.8249f); // Reference white Adobe Photoshop
	
	float X = xr * white.x;
	float Y = yr * white.y;
	float Z = zr * white.z;	

	return float3(X, Y, Z);
}

float InvGammaCorrection(float value)
{
	float ret;
	if (value <= 0.0031308f) {
		ret = value*12.92f;
	} else {
		ret = 1.055f * pow(value,1.0f/2.4f) - 0.055f;
	}
	return ret;
}

float3 InvRGBGammaCorrection(float3 rgbcol)
{
	float X = InvGammaCorrection(rgbcol.x);
	float Y = InvGammaCorrection(rgbcol.y);
	float Z = InvGammaCorrection(rgbcol.z);
	return float3(X,Y,Z);
}

////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader
////////////////////////////////////////////////////////////////////////////////////

void SpriteVertexShader(inout float4 color    : COLOR0,

                       inout float2 texCoord : TEXCOORD0,

                       inout float4 position : POSITION0)

{

   // Half pixel offset for correct texel centering.

   position.xy -= 0.5f;

   // Viewport adjustment.

   position.xy = position.xy / viewport;

   position.xy *= float2(2.0f, -2.0f);

   position.xy -= float2(1.0f, -1.0f);

}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 ConvertPS(VertexShaderOutput input) : COLOR0
{
	
	//float2 texC = input.pos.xy /= input.pos.w;
	//texC.x =  0.5f*texC.x + 0.5f; 
	//texC.y = -0.5f*texC.y + 0.5f;   //needed?
	
	//float3 color = float3(0.0f, 0.0f, 0.0f);
	//float3 black = float3(0.0f, 0.0f, 0.0f);

	// get color from texture
	float4 color = tex2D(labtex, input.texC);

	if(true)//(color.x > 0.05f) || (color.y > 0.05f) || (color.z > 0.05f))
	{
		// convert from [0,+1] to [0,510] and shift to [-255,255]:
		color.xyz = float3(color.xyz*510.0f)-255.0f;

		// rotate according to specified color vision deficiency, rotate plane back
		color.xyz = mul(float4(color.xyz, 1.0f), rotmat).xyz;
		
		// convert from L*a*b* back to XYZ
		color.xyz = LabToXYZ(color.xyz);
					
		// convert from XYZ back to RGB
		color = mul(color, mXYZtoRGB);

		// do inverse gamma correction
		color.xyz = InvRGBGammaCorrection(color.xyz); 

		// scale to [0 1]
		//color.xyz = (color.xyz+255.0f)/510.0f; 

	}

	color.w = 0.0f;
	return color;
}

technique LabtoRGB
{
    pass Pass1
    {	
		VertexShader = compile vs_3_0 SpriteVertexShader();	
        PixelShader = compile ps_3_0 ConvertPS();
    }
}