//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
texture2D redTexture;
//float2 dim;
float2 viewport;

//Samplers -------------------------------------------------------------------------

sampler2D redtex = sampler_state
{
	Texture = <redTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};


////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader
////////////////////////////////////////////////////////////////////////////////////

void SpriteVertexShader(inout float4 color    : COLOR0,

                       inout float2 texCoord : TEXCOORD0,

                       inout float4 position : POSITION0)

{

   // Half pixel offset for correct texel centering.

   position.xy -= 0.5f;

   // Viewport adjustment.

   position.xy = position.xy / viewport;

   position.xy *= float2(2.0f, -2.0f);

   position.xy -= float2(1.0f, -1.0f);

}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 PCA4PS(PixelShaderInput input) : COLOR0
{
	// compute max of projection effect

	float4 color = float4(0.0f, 0.0f, 0.0f, 0.0f);

	if((input.texC.x*viewport.x < 1.0f) && (input.texC.y*viewport.y < 1.0f))
	{
		for(int i = 0.0f; i<viewport.x; i++)
		{
			for(int j = 0.0f; j<viewport.y; j++)
			{
				float xCoord = i/viewport.x;
				float yCoord = j/viewport.y;
				color = max(abs(color), abs(tex2Dlod(redtex, float4(xCoord,yCoord,0.0f,0.0f))-0.5f)); 
			}
		}
	}	

	// shift [0, 0.5] to [0 1] ... abs -> all positive
	color.xyz = color.xyz+0.5f; 
	
	if((color.x > 1.0f) || (color.y > 1.0f) || (color.z > 1.0f))
	{
		color.xyz = float3(0.0f, 0.0f, 1.0f);
	}
	else if((color.x < 0.0f) || (color.y < 0.0f) || (color.z < 0.0f))
	{
		color.xyz = float3(0.0f, 1.0f, 0.0f);
	}
	
	color.w = 1.0f;
	return color;
}

technique PCA4
{
    pass Pass1
    {		
		VertexShader = compile vs_3_0 SpriteVertexShader();
        PixelShader = compile ps_3_0 PCA4PS();
    }
}



