//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------
texture2D redTexture;
//float2 dim;
float2 viewport;

//Samplers -------------------------------------------------------------------------

sampler2D redtex = sampler_state
{
	Texture = <redTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader
////////////////////////////////////////////////////////////////////////////////////

void SpriteVertexShader(inout float4 color    : COLOR0,

                       inout float2 texCoord : TEXCOORD0,

                       inout float4 position : POSITION0)

{

   // Half pixel offset for correct texel centering.

   position.xy -= 0.5f;

   // Viewport adjustment.

   position.xy = position.xy / viewport;

   position.xy *= float2(2.0f, -2.0f);

   position.xy -= float2(1.0f, -1.0f);

}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 PCA2PS(PixelShaderInput input) : COLOR0
{
	// input are weighted directions of contrast loss.
	// compute predominant component for whole image: sum up all weighted directions
	
	float4 color = float4(0.0f, 0.0f, 0.0f, 0.0f);

	if((input.texC.x*viewport.x < 1.0f) && (input.texC.y*viewport.y < 1.0f))
	{
		//color = tex2D(redtex, input.texC)-0.5f;  // shift!

		for(int i = 0.0f; i<viewport.x; i++)
		{
			for(int j = 0.0f; j<viewport.y; j++)
			{
				float xCoord = i/viewport.x;
				float yCoord = j/viewport.y;

				color = color + tex2Dlod(redtex, float4(xCoord,yCoord,0.0f,0.0f))-0.5f;
			}
		}

		// scale to [-1 1]
		float maxC = 0.0f;
		if(abs(color.x) > abs(color.y))
		{ 
			maxC = abs(color.x);
			if(abs(color.z) > maxC ){ maxC = abs(color.z);}
		} 
		else 
		{
			maxC = abs(color.y);
			if(abs(color.z) > maxC ){ maxC = abs(color.z);}
		}
		color = color/maxC;
		
		//float colorLength = sqrt(pow(color.x,2.0f)+pow(color.y,2.0f)+pow(color.z,2.0f));
		//color = color/colorLength;

		// scale from [-1 1] to [-0.5 0.5]
		color = color/2.0f;
	}	

	// shift [-0.5, 0.5] to [0 1]
	color.xyz = color.xyz+0.5f; 
	
	if((color.x > 1.0f) || (color.y > 1.0f) || (color.z > 1.0f))
	{
		color.xyz = float3(0.0f, 0.0f, 1.0f);
	}
	else if((color.x < 0.0f) || (color.y < 0.0f) || (color.z < 0.0f))
	{
		color.xyz = float3(0.0f, 1.0f, 0.0f);
	}
	
	color.w = 1.0f;
	//color.xy = input.texC;
	return color;
}

technique PCA2
{
    pass Pass1
    {		
		VertexShader = compile vs_3_0 SpriteVertexShader();
        PixelShader = compile ps_3_0 PCA2PS();
    }
}



