//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------

texture2D eigenTexture;
texture2D labTexture;

//Samplers -------------------------------------------------------------------------

sampler2D eigentex = sampler_state
{
	Texture = <eigenTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

sampler2D labtex = sampler_state
{
	Texture = <labTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
    //BorderColor = float4(0,0,0,0);	// outside of border should be black
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};

//----------------------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader
////////////////////////////////////////////////////////////////////////////////////

float4 ProjPS(PixelShaderInput input) : COLOR0
{
	/*orthographic projection of the original colors
	onto the plane defined by the vectors L and vab (=main EigenVector) */

	// get main EigenVector for whole image at position (0,0) and shift from [0 1] to [-0.5 0.5]
	float4 vab = tex2D(eigentex, float2(0.0f,0.0f))-0.5;

	float4 color = tex2D(labtex, input.texC)-0.5;

	//shift from [0 1] to [-0.5 0.5]
	//color.xyz = color.xyz -0.5f;

	//shift from [0 1] to actuel values
	//color.xyz = color.xyz *10000.0f-2000.0f;

	// convert from [0,+1] to [0,510] and shift to [-255,255]:
	//color.xyz = (color.xyz*510.0f)-255.0f;

	float L = color.x;
	float a = color.y;
	float b = color.z;
	float va = vab.x;
	float vb = vab.y;

	float3 projColor = (0.0f,0.0f,0.0f);
	float4 outColor = (0.0f,0.0f,0.0f,color.w);

	// L*:
	outColor.x = L; // not really needed
	// a* = 0.0f; but save here length of ab to enhance contrast later
	outColor.y = sqrt(pow(a,2.0f)+pow(b,2.0f));
	// b* = length of ab projected on vab = dot(ab,vab)
	outColor.z = a * va + b * vb; // dot product 

	// project to plane
	// C_proj = k * vab
	/*float k = (a*va + b*vb) / (pow(va,2.0f) + pow(vb,2.0f));
	projColor.y = k * va;
	projColor.z = k * vb;*/

	//outColor.x = a * va + b * vb; // dot product ->projection effect
	//outColor.y = sqrt(pow(a,2.0f)+pow(b,2.0f)); // lentgh of ab to enhance contrast

	// rotate around L* onto a*-plane 
	// a* =||C_proj.ab - Origin.ab|| = sqrt( a_proj^2 + b_proj^2 ) 
	// => a* = sqrt( a_proj^2 + b_proj^2 ) ... y
	// b* = 0 ... z
	// L* = L ... x
	//outColor.x = L;
	//outColor.y = sqrt(pow(projColor.y,2.0f)+pow(projColor.z,2.0f));
	//outColor.z = 0.0f;	

	//onto b* plane:
	/*outColor.x = L;
	outColor.y = 0.0f;
	outColor.z = sqrt(pow(projColor.y,2.0f)+pow(projColor.z,2.0f));*/
	
	// scale to [0 1]
	//outColor.xyz = (outColor.xyz+2000.0f)/10000.0f;

	//outColor.xyz = outColor.xyz+0.5f; 

	// shift [-0.5, 0.5] to [0 1]
	outColor.xyz = outColor.xyz+0.5f; 

	/*if((outColor.z <= 0.5f))
	{
		outColor.xyz = float3(0.0f, 1.0f, 0.0f);
	}
	else
	{
		outColor.xyz = float3(0.0f, 0.0f, 1.0f);
	}*/

	/*if((outColor.x < 0.0f) || (outColor.y < 0.0f) || (outColor.z < 0.0f))
	{
		outColor.xyz = float3(0.0f, 0.0f, 1.0f);
	}

	if((outColor.x > 1.0f) || (outColor.y > 1.0f) || (outColor.z > 1.0f))
	{
		outColor.xyz = float3(0.0f, 1.0f, 0.0f);
	}*/
	
	//outColor = vab+0.5;	
	outColor.w = 1.0f;	
	return outColor;
}


technique Project
{
    pass Pass1
    {		
        PixelShader = compile ps_2_0 ProjPS();
    }
}

