//----------------------------------------------------------------------------------
//Global variables
//----------------------------------------------------------------------------------

float4x4 rotmat; 
float2 viewport;
texture2D RGBtexture;

// RGB to XYZ matrix conversion 
float4x4 mRGBtoXYZ;

//Samplers -------------------------------------------------------------------------

sampler2D rgbtex = sampler_state
{
	Texture = <RGBtexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	
	AddressU = CLAMP;				// border sampling in U
    AddressV = CLAMP;				// border sampling in V
};

//----------------------------------------------------------------------------------

// Pixel Shader input structure (= Vertex Shader Output)
struct VertexShaderOutput
{
    float2 texC			: TEXCOORD0;
};

// Pixel Shader input structure (= Vertex Shader Output)
struct PixelShaderInput
{
    float2 texC			: TEXCOORD0;
};

struct PixelShader1Output
{
    float4 color			: COLOR0;
};

//----------------------------------------------------------------------------------

float F(float p)
{
	if (p>0.008856f)
	{
		return pow(p,1.0f/3.0f);
	}
	else
	{
		return (p*903.3f + 16.0f)/116.0f;
	}	
}

float3 XYZToLab(float3 vec3)
{
	// divide input (X, Y, Z) by reference white
	float3 white = ( 0.9642f, 1.0f, 0.8249f); // Reference white Adobe Photoshop

	float xr = vec3.x/white.x;
	float yr = vec3.y/white.y;
	float zr = vec3.z/white.z;

	float fx = F(xr);
	float fy = F(yr);
	float fz = F(zr);

	float L = 116.0f * fy - 16.0f;
	float a = 500.0f * (fx - fy);
	float b = 200.0f * (fy - fz);

	return float3(L, a, b);
}

float gammaCorrection(float value)
{
	float ret;
	if (value <= 0.04045f) {
		ret = value/12.92f;
	} else {
		ret = pow(((value + 0.055f)/1.055f),2.4f);
	}
	return ret;
}

float3 RGBGammaCorrection(float3 rgbcol)
{
	float X = gammaCorrection(rgbcol.x);
	float Y = gammaCorrection(rgbcol.y);
	float Z = gammaCorrection(rgbcol.z);
	float3 result = float3(X,Y,Z);
	return result;
}

////////////////////////////////////////////////////////////////////////////////////
//Vertex Shader
////////////////////////////////////////////////////////////////////////////////////

void SpriteVertexShader(inout float4 color    : COLOR0,

                       inout float2 texCoord : TEXCOORD0,

                       inout float4 position : POSITION0)

{

   // Half pixel offset for correct texel centering.

   position.xy -= 0.5f;

   // Viewport adjustment.

   position.xy = position.xy / viewport;

   position.xy *= float2(2.0f, -2.0f);

   position.xy -= float2(1.0f, -1.0f);

}

////////////////////////////////////////////////////////////////////////////////////
//Pixel Shader 
////////////////////////////////////////////////////////////////////////////////////

float4 ConvertPS(PixelShaderInput input) : COLOR0 
{
	// get color from texture
	float4 color = tex2D(rgbtex, input.texC);

	if(true)//(color.x > 0.05f) || (color.y > 0.05f) || (color.z > 0.05f))
	{
	
	// do gamma correction
	color.xyz = RGBGammaCorrection(color.xyz); //checked! makes picture a little bit brighter

	// convert from RGB to XYZ
	color = mul(color, mRGBtoXYZ); //checked!
				
	// convert from XYZ to Lab
	color.xyz = XYZToLab(color.xyz);	
		
	// rotate according to specified color vision deficiency
	color.xyz = mul(float4(color.xyz, 1.0f), rotmat).xyz;

	// shift from [-255,255] to [0,510] and convert to [0,+1]:
	color.xyz = (color.xyz+255.0f)/510.0f; 

	if((color.x > 1.0f) || (color.y > 1.0f) || (color.z > 1.0f))
	{
		color.xyz = float3(0.0f, 0.0f, 1.0f);
	}
	else if((color.x < 0.0f) || (color.y < 0.0f) || (color.z < 0.0f))
	{
		color.xyz = float3(0.0f, 1.0f, 0.0f);
	}

	}
	
	color.w = 0.0f;		
	return color;
}

technique RGBtoLab
{
    pass Pass1
    {		
		VertexShader = compile vs_3_0 SpriteVertexShader();
        //VertexShader = compile vs_3_0 ConvertVS();
		PixelShader = compile ps_3_0 ConvertPS();
    }
}