﻿#region Using Statements
using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using Microsoft.Xna.Framework.Graphics.PackedVector;
using visLU.Effects;
using visLU.CVD;
#endregion

namespace visLU
{
    /// <summary>
    /// Render engine.
    /// Implenets a Windows Forms Control
    /// </summary>
    class Engine : VisXnaConrol
    {
        #region Variables

        #region Render Targets
        private RenderTarget2D frontTarget;
        private RenderTarget2D backTarget;
        private RenderTarget2D sceneTarget;
        private RenderTarget2D dvrTarget;

        private Texture2D dvrTexture;

        private RenderTarget2D cvdTarget;
        #endregion

        ContentManager content;
        SpriteBatch spriteBatch;
        SpriteFont spriteBatchFont;

        #region viewport
        Camera[] cameras;
        Camera cvdCamera;
        Viewport defaultViewport;
        Viewport[] viewports;
        int currentView;
        int currentCvdView;
        #endregion

        Data input;
        private Model cubeModel;

        #region shader

        //standad
        public VolumeShader shader;

        // CVD
        public VolumeShader CVDshader;
        public TFWeightsShader tfWeightsShader;

        // recoloring
        public RecolorShader rGBtoLab_Shader;
        public RecolorShader contrastLoss_Shader;
        public RecolorShader predominantComponent_Shader;
        public RecolorShader eigenVector_Shader;
        public RecolorShader projection_Shader;
        public RecolorShader maxProjectionEffect_Shader;
        public RecolorShader enhanceContrast_Shader;
        public RecolorShader labToRGB_Shader;

        //simulation
        public RecolorShader sim_RGBtoLab_Shader;
        public RecolorShader lMSShader;
        public RecolorShader sim_LabToRGB_Shader;

        #endregion

        Image recoloredImage;
        Image simulatedImage;
        CVDtf cvdtf;


        //int texturesSaved = 0; // for debug
        int render = 3; //0 = nothing, 1=frontfaces, 2= front and backfaces 3= + volume
        bool debugOutput = GameProperties.Instance.debugOutput;
        int drawcount = 0; //only for debug       

        int[] lastSize;

        Texture2D frameworkInfo;

        private bool debugShader;
        Texture2D rgbTestTex;

        #endregion

        #region Properties
        public Data Input
        {
            set { input = value; }
            get { return input; }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// Creates data and 4 cameras (one for each view: 3D, front, top, side).
        /// </summary>
        /// <param name="_game"></param>
        public Engine()
        {
            #region create cameras
            cameras = new Camera[4];
            cameras[0] = new InteractiveCamera();
            cameras[0].Initialize();
            for (int i = 1; i < 4; i++)
            {
                cameras[i] = new Camera();
                cameras[i].Initialize();
            }
            cvdCamera = new InteractiveCamera(); cvdCamera.Initialize();
            currentCvdView = (int)GameProperties.Instance.viewMode;
            currentCvdView = (int)GameProperties.Instance.cvdViewMode; //start view for the split screen
            #endregion

            input = null;
            lastSize = new int[] {this.Size.Width, this.Size.Height};

            debugShader = GameProperties.Instance.shaderDebugMode;

            if (debugOutput)
            {
                Console.WriteLine("ENGINE STARTED");
            }

        }
        #endregion

        #region Methods
        //##########################################################################################

        #region Initialize()
        //-----------------------------------INITIALIZE()-------------------------------------------

        /// <summary>
        /// Allows the game component to perform any initialization it needs to before starting
        /// to run.  This is where it can query for any required services and load content.
        /// Initialzes cameras with position, target and up-vector.
        /// </summary>
       protected override void Initialize()
        {   
            //get the content manager
            content = Services.GetService(typeof(ContentManager)) as ContentManager;
            if (content == null)
            {
                content = new ContentManager(Services, "Content");
                Services.AddService<ContentManager>(content);
            }

            //init sprite batch
            spriteBatch = new SpriteBatch(GraphicsDevice);
            Services.AddService<SpriteBatch>(spriteBatch);

            #region set views

            //cameras[1] -> top slice view
            cameras[1].CameraPosition = new Vector3(0.0f, 1.0f, 0.0f);
            cameras[1].CameraTarget = GameProperties.Instance.dataOrigin;
            cameras[1].CameraUpVector = Vector3.UnitZ;
            cameras[1].SetPerspectiveProjecton = false;
            cameras[1].Initialize();

            //cameras[2] -> side slice view
            cameras[2].CameraPosition = new Vector3(1.0f, 0.0f, 0.0f);
            cameras[2].CameraTarget = GameProperties.Instance.dataOrigin;
            cameras[2].CameraUpVector = Vector3.Up;
            cameras[2].SetPerspectiveProjecton = false;
            cameras[2].Initialize();

            //cameras[3] -> front slice view
            cameras[3].CameraPosition = new Vector3(0.0f, 0.0f, 1.0f);
            cameras[3].CameraTarget = GameProperties.Instance.dataOrigin;
            cameras[3].CameraUpVector = Vector3.Up;
            cameras[3].SetPerspectiveProjecton = false;
            cameras[3].Initialize();

            #endregion

            #region init control points
            setInitialControlPoints();
            #endregion
        }
        //__________________________________________________________________________________________
        #endregion

        #region LoadContent()
        //-----------------------------------LOAD_CONTENT()-----------------------------------------
        /// <summary>
        /// Creates spriteBatch to display current view name.
        /// Calls initializeTextureFilter().
        /// </summary>
        protected override void LoadContent()
        {
            spriteBatchFont = content.Load<SpriteFont>("Fonts\\spriteFont1");

            frameworkInfo = content.Load<Texture2D>("Texture\\framework_info");

            if (debugShader)
            {
                //rgbTestTex = content.Load<Texture2D>("Texture\\rgb");
                //rgbTestTex = content.Load<Texture2D>("Texture\\test1");
                rgbTestTex = content.Load<Texture2D>("Texture\\schiff");
                //rgbTestTex = content.Load<Texture2D>("Texture\\nebula");
                //rgbTestTex = content.Load<Texture2D>("Texture\\lobster");
            }

            #region viewports
            //init viewports for split screen
            defaultViewport = GraphicsDevice.Viewport;
            defaultViewport.Width = this.Size.Width;
            defaultViewport.Height = this.Size.Height;
            setViewports();
            #endregion

            #region create Render Targets

            PresentationParameters pp = GraphicsDevice.PresentationParameters;
            SurfaceFormat format = pp.BackBufferFormat;
            //format = SurfaceFormat.Rgba64;
            int width = pp.BackBufferWidth;
            int height = pp.BackBufferHeight;
            DepthFormat depthStecilFormat = pp.DepthStencilFormat;
            int multiSample = pp.MultiSampleCount;


            frontTarget = new RenderTarget2D(GraphicsDevice, width, height, false, format,
                depthStecilFormat, multiSample, RenderTargetUsage.DiscardContents);

            backTarget = new RenderTarget2D(GraphicsDevice, width, height, false, format,
                depthStecilFormat, multiSample, RenderTargetUsage.DiscardContents);

            sceneTarget = new RenderTarget2D(GraphicsDevice, width, height, false, format,
                depthStecilFormat, multiSample, RenderTargetUsage.DiscardContents);

            dvrTarget = new RenderTarget2D(GraphicsDevice, width, height, false, format, depthStecilFormat, multiSample, RenderTargetUsage.DiscardContents);

            cvdTarget = new RenderTarget2D(GraphicsDevice, width, height, false, SurfaceFormat.Rgba64, depthStecilFormat, multiSample, RenderTargetUsage.DiscardContents);

               
      

            Console.WriteLine("RENDERTARGETS created!");
            #endregion

            #region create all shader

            if (debugOutput)
            {
                Console.WriteLine("trying to create shader");
            }

            // standard shader
            shader = new VolumeShader(GraphicsDevice, content.Load<Effect>("Effects\\volumeShader"), false);

            #region CVD shader

            CVDshader = new VolumeShader(GraphicsDevice, content.Load<Effect>("Effects\\volumeCVDShader"), true);
            tfWeightsShader = new TFWeightsShader(GraphicsDevice, content.Load<Effect>("Effects\\TFWeights"));

            #endregion

            #region recolor shader

            rGBtoLab_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\RGBtoLab"), "convertToLab");
            contrastLoss_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\ContrastLoss"), "ContrastLoss");
            predominantComponent_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\PredominantComponent"), "PredominantComponent");
            eigenVector_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\EigenVector"), "EigenVector");
            projection_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\Proj"), "Project");
            maxProjectionEffect_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\MaxProjectionEffect"), "MaxProjectionEffect");
            enhanceContrast_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\EnhanceContrast"), "EnhanceContrast");
            labToRGB_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\LabtoRGB"), "convertToRGB");            

            #endregion

            #region simulation shader

            sim_RGBtoLab_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\RGBtoLab"), "convertToLab");
            lMSShader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\LMS"), "LMS");
            sim_LabToRGB_Shader = new RecolorShader(GraphicsDevice, content.Load<Effect>("Effects\\LabtoRGB"), "convertToRGB");
            
            #endregion

            #endregion

            cubeModel = content.Load<Model>("Models/box");

        }


        //__________________________________________________________________________________________
        #endregion

        #region UnloadContent()
        //-----------------------------------UNLOAD_CONTENT()---------------------------------------

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        /*protected override void UnloadContent() ->>> free resources???
        {
            // TODO: Unload any non ContentManager content here
            base.UnloadContent();
        }*/

        //__________________________________________________________________________________________
        #endregion

        #region Update()
        //-----------------------------------UPDATE()-----------------------------------------------

        /// <summary>
        /// Allows the game component to update itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update()
        {

            if (debugOutput)
            {
                Console.WriteLine("Engine Update Call number " + drawcount);
            }
            
            #region load data

            if (GameProperties.Instance.engineState.loadData)
            {
               
                if (input != null && this.Controls.Count > 0)
                {
                    foreach (Control c in Controls)
                    {
                        if (c.GetType().Equals("Data"))
                        {
                            this.Controls.Remove(c);
                            input = null;
                            break;
                        }

                    }
                }
                setInitialControlPoints();
                input = new Data(GameProperties.Instance.dataDir + GameProperties.Instance.dataFilename);
                this.Controls.Add(input);

                #region init camera for 3d view
                //init camera
                 Vector4 factor = input.scaleFactor;
                 Vector3 scale = new Vector3(factor.X, factor.Y, factor.Z);
                 Vector3 target = new Vector3(0.5f, 0.5f, 0.5f) * scale;
                 target = Vector3.Transform(target, input.worldMatrix);
                 cameras[0] = new InteractiveCamera();
                 cameras[0].CameraPosition = new Vector3(target.X, target.Y, 2.0f);
                 cameras[0].CameraTarget = target;
                 cameras[0].Initialize();
                 cvdCamera = new InteractiveCamera();
                 cvdCamera.CameraPosition = new Vector3(target.X, target.Y, 2.0f);
                 cvdCamera.CameraTarget = target;
                 cvdCamera.Initialize();
                #endregion

                 shader.SetEffectParameter(input.VolumeData, input.volumeGradients, GameProperties.Instance.directLightPos, input.stepSize * input.scaleSteps, GameProperties.Instance.baseSampleDist, input.scaleSteps, input.iterations, 0, input.scaleFactor, new Vector2(input.DataDimensions.X, input.DataDimensions.Y), input.slice);
                //shader.SetEffectParameter(input.VolumeData, input.volumeGradients, input.stepSize * input.scaleSteps, input.iterations, 0, input.scaleFactor, input.slice);
                GameProperties.Instance.engineState.loadData = false;
                GameProperties.Instance.showInfoText = false;
                
                if (debugOutput)
                {
                    Console.WriteLine("load data done");
                }

            } //end if load data
            #endregion

            #region set view mode and camera

            //set view mode and camera for single screen
            if (GameProperties.Instance.engineState.updateCamera)
            {
                #region set camera to current view
                if (debugOutput)
                {
                    Console.WriteLine("SETTING CAMERA");
                }
                if (GameProperties.Instance.viewportMode.Equals(ViewportMode.Single))
                {
                    currentView = 0;
                    switch (GameProperties.Instance.viewMode)
                    {
                        case ViewMode.VolView: //single 3d view

                            currentView = 0;
                            break;

                        case ViewMode.TopView: //single top view
                            currentView = 1;
                            break;

                        case ViewMode.SideView: //single side view
                            currentView = 2;
                            break;

                        case ViewMode.FrontView: //single front view
                            currentView = 3;
                            break;

                        default:
                            currentView = 0;
                            break;
                    }

                    if (currentView == 0) //set camera for the 3d view
                    {
                        cameras[currentView].SetPerspectiveProjecton = GameProperties.Instance.perspectiveProjection;
                    }

                    cameras[currentView].CameraInput = GameProperties.Instance.cameraInput;
                    cameras[currentView].Update();


                #endregion

                    //set parameters to draw the correct scene view
                    if (input != null)
                    {
                        input.Camera = cameras[currentView];
                        input.CurrentCamera = currentView;
                    }
                }
                else if (GameProperties.Instance.viewportMode.Equals(ViewportMode.Split4))//set view mode and camera vor split screen
                {
                    //3d view camera
                    cameras[0].SetPerspectiveProjecton = GameProperties.Instance.perspectiveProjection;
                    cameras[0].CameraInput = GameProperties.Instance.cameraInput;
                    cameras[0].SetClientSize = new int[] { viewports[0].Width, viewports[0].Height };
                    cameras[0].Update();
                    cvdCamera.SetPerspectiveProjecton = GameProperties.Instance.perspectiveProjection;
                    cvdCamera.CameraInput = GameProperties.Instance.cameraInput;
                    cvdCamera.SetClientSize = new int[] { viewports[2].Width, viewports[2].Height };
                    cvdCamera.Update();

                    if (GameProperties.Instance.enableCVDView == true) //priority!
                    {
                        cvdCamera.SetPerspectiveProjecton = GameProperties.Instance.perspectiveProjection;
                        cvdCamera.CameraInput = GameProperties.Instance.cameraInput;
                        cvdCamera.SetClientSize = new int[] { viewports[2].Width, viewports[2].Height };
                        cvdCamera.Update();
                    }
                    else  //split screen with 3d and slices 
                    {
                        for (int i = 1; i < cameras.Length; i++)
                        {
                            cameras[i].CameraInput = GameProperties.Instance.cameraInput;
                            cameras[i].SetClientSize = new int[] { viewports[i].Width, viewports[i].Height };
                            cameras[i].Update();
                        }
 
                    }
                }

                GameProperties.Instance.engineState.updateCamera = false;
                if (debugOutput)
                {
                    Console.WriteLine("CAMERA SET");
                }
            } 
            #endregion
         
            #region update data
            if (GameProperties.Instance.engineState.updateData && input != null)
            {
                #region front and backfaces

                if (GameProperties.Instance.viewportMode.Equals(ViewportMode.Split4) || (GameProperties.Instance.viewportMode.Equals(ViewportMode.Single) && currentView == (int)ViewMode.VolView))
                {   
                    #region set shader parameters for rendering
                    GraphicsDevice.BlendState = BlendState.Additive;
                    shader.CurrentTechnique = shader.Techniques["RenderPosition"];
                    shader.SetEffectParameter(input.worldMatrix * cameras[(int)ViewMode.VolView].ViewProjection, input.worldMatrix);
                    shader.SetEffectParameter(cameras[(int)ViewMode.VolView].CameraPosition);
                    #endregion

                    if (render == 0)
                    {
                        if (debugOutput)
                        {
                            Console.WriteLine("don't draw frontfaces or backfaces or volume");
                        }
                    }

                    if (render > 0)
                    {
                        #region Draw Frontfaces
                        if (debugOutput)
                        {
                            Console.WriteLine("1 DRAWING FRONTFACES");
                        }
                        GraphicsDevice.SetRenderTarget(frontTarget);
                        GraphicsDevice.Clear(Color.Black);
                        GraphicsDevice.RasterizerState = RasterizerState.CullCounterClockwise;
                        shader.FrontFacesTexture = null;

                        DrawVolume(0);
                        GraphicsDevice.SetRenderTarget(null);
                        #endregion
                    }

                    if (render > 1)
                    {
                        #region Draw Backfaces
                        if (debugOutput)
                        {
                            Console.WriteLine("2 DRAWING BACKFACES");
                        }
                        GraphicsDevice.SetRenderTarget(backTarget);
                        GraphicsDevice.Clear(Color.Black);
                        GraphicsDevice.RasterizerState = RasterizerState.CullClockwise;
                        shader.BackFacesTexture = null;

                        DrawVolume(0);
                        GraphicsDevice.SetRenderTarget(null);
                        #endregion
                    }

                    #region pass textures to shader
                    shader.FrontFacesTexture = frontTarget;
                    shader.BackFacesTexture = backTarget;
                    #endregion

                    input.setControlpoints(); 
                    shader.TransferTexture = input.transferTexture;
                    shader.TransferTextureAlpha = input.transferTextureAlpha; 
                }

                
                #region pass textures to shader
                shader.FrontFacesTexture = frontTarget;
                CVDshader.FrontFacesTexture = frontTarget;
                shader.BackFacesTexture = backTarget;
                CVDshader.BackFacesTexture = backTarget;
                #endregion

                input.setControlpoints();
                shader.TransferTexture = input.transferTexture; 
                shader.TransferTextureAlpha = input.transferTextureAlpha; 
                CVDshader.TransferTextureAlpha = input.transferTextureAlpha; 

               GameProperties.Instance.engineState.updateData = false;

                if (debugOutput)
                {
                    Console.WriteLine("data updated");
                }
            }
            #endregion
            #endregion

            #region optimize tf for cvd
            //process just once for a single tf
            if (true/*!(GameProperties.Instance.recolor) && drawcount > 5*/) //just for debug
            {
                if (GameProperties.Instance.enableCVDView && input != null)
                {

                    #region recoloring

                    #region calculate new?
                    // calculate new only when recoloredView is first activated
                    if (GameProperties.Instance.recoloredView == true && GameProperties.Instance.recolor == false && GameProperties.Instance.triggerRecolor)
                    {
                        GameProperties.Instance.recolor = true;
                    }
                    else
                    {
                        GameProperties.Instance.recolor = false;
                    }
                    #endregion

                    if (GameProperties.Instance.recolor)
                    {
                        #region instanciate Image class
                        if (recoloredImage == null)
                        {
                            recoloredImage = new Image(GraphicsDevice, GameProperties.Instance.CVD);
                        }
                        #endregion

                        #region assign shader
                        if (recoloredImage.RGBtoLab_Shader == null)
                        {
                            recoloredImage.RGBtoLab_Shader = rGBtoLab_Shader;
                            recoloredImage.ContrastLoss_Shader = contrastLoss_Shader;
                            recoloredImage.PredominantComponent_Shader = predominantComponent_Shader;
                            recoloredImage.EigenVector_Shader = eigenVector_Shader;
                            recoloredImage.Projection_Shader = projection_Shader;
                            recoloredImage.MaxProjectionEffect_Shader = maxProjectionEffect_Shader;
                            recoloredImage.EnhanceContrast_Shader = enhanceContrast_Shader;
                            recoloredImage.LabToRGB_Shader = labToRGB_Shader;
                        }
                        #endregion
                        
                        #region debug: set test image to recolor
                        if (debugShader)
                        {
                            recoloredImage.ReferenceImageRGB = rgbTestTex;
                        }
                        else
                        {
                            if (GameProperties.Instance.viewportMode.Equals(ViewportMode.Split4))
                            {
                                recoloredImage.ReferenceImageRGB = dvrTexture;
                                //saveScreenshot(dvrTarget, "dvrtest.png", "PNG");
                            }
                            else
                            {
                                recoloredImage.ReferenceImageRGB = sceneTarget;
                            }
                        }
                        #endregion

                        // recolor image
                        recoloredImage.Recolor(GraphicsDevice);

                        GameProperties.Instance.recolor = false;
                        GameProperties.Instance.triggerRecolor = false;
                        
                    }

                    #endregion

                    #region calculate cvd tf

                    if (recoloredImage != null  && GameProperties.Instance.cvdView && GameProperties.Instance.triggerOptimize)
                    {
                        cvdtf = new CVDtf(recoloredImage.RecoloredImage, GraphicsDevice);
                        cvdtf.selectRecoloredPixels();

                        #region set shader
                        //set selected pixels to shader
                        tfWeightsShader.SetEffectParameter(cvdtf.RecoloredImageMask);
                        tfWeightsShader.FrontFacesTexture = shader.FrontFacesTexture;
                        tfWeightsShader.BackFacesTexture = shader.BackFacesTexture;
                        tfWeightsShader.TransferTexture = shader.TransferTexture;
                        tfWeightsShader.TransferTextureAlpha = shader.TransferTextureAlpha;
                        tfWeightsShader.SetEffectParameter(input.volumeData, GameProperties.Instance.directLightPos, input.stepSize * input.scaleSteps, GameProperties.Instance.baseSampleDist, input.scaleSteps, input.iterations, 0, input.scaleFactor, new Vector2(input.DataDimensions.X, input.DataDimensions.Y), input.slice);
                        tfWeightsShader.SetEffectParameter(input.worldMatrix * cameras[0].ViewProjection, input.worldMatrix);
                        tfWeightsShader.SetEffectParameter(cameras[0].CameraPosition);
                        tfWeightsShader.CurrentTechnique = tfWeightsShader.Techniques["CalculateColorWeights"];
                        #endregion

                        #region set render graphics device
                        //GraphicsDevice.DepthStencilState = DepthStencilState.Default;
                        GraphicsDevice.BlendState = BlendState.AlphaBlend;
                        GraphicsDevice.RasterizerState = RasterizerState.CullNone;
                        #endregion

                        #region calculate weights

                        List<TransferControlPoint> cpList = GameProperties.Instance.colorControlPoints;
                        cpList.Sort(delegate(TransferControlPoint p1, TransferControlPoint p2) { return p1.isoValue.CompareTo(p2.isoValue); });

                        //render for each control point
                        for (int i = 0; i < cpList.Count; i++)
                        {

                            float currCp; float prevCp; float nextCp;
                            //set curr cp index to shader
                            if (i == 0)
                            {
                                currCp = (float)cpList[i].isoValue;
                                prevCp = 0.0f;//1. default cp
                                nextCp = (float)cpList[i + 1].isoValue;
                            }
                            else if (i == (cpList.Count - 1))
                            {
                                currCp = (float)cpList[i].isoValue;
                                prevCp = (float)cpList[i - 1].isoValue;
                                nextCp = 255.0f; //last default cp
                            }
                            else
                            {
                                currCp = (float)cpList[i].isoValue;
                                prevCp = (float)cpList[i - 1].isoValue;
                                nextCp = (float)cpList[i + 1].isoValue;
                            }
                            tfWeightsShader.SetEffectParameter(currCp, prevCp, nextCp);
                            #region draw
                            GraphicsDevice.SetRenderTarget(cvdTarget); //cvdTarget
                            GraphicsDevice.Clear(Color.Black);
                            if (GameProperties.Instance.debugOutput)
                            {
                                Console.WriteLine("GraphicsDevice Render Target set to cvd target");
                            }
                            //call draw with this shader, set all other parameteres needed by the ray casting
                            try
                            {
                                foreach (ModelMesh mesh in cubeModel.Meshes)
                                {
                                    foreach (ModelMeshPart mmp in mesh.MeshParts)
                                    {
                                        mmp.Effect = tfWeightsShader;

                                        Console.WriteLine("drawing...side.." + currentView);
                                    };
                                    mesh.Draw();
                                }
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine("ERROR: " + e);
                                throw;
                            }

                            GraphicsDevice.SetRenderTarget(null);
                            #endregion
                            cvdtf.setTFWeights(cvdTarget, i); //cvdTarget for ith control point
                            if (debugOutput)
                            {
                                saveScreenshot(cvdTarget, "cpweights" + i, "png");
                            }
                        }//end for control points list

                        #endregion

                        #region calculate and set cvd transfer function
                        CVDshader.CVDTransferTexture = cvdtf.calculateCVDColors();
                        #endregion

                        GameProperties.Instance.CVDTFOptimized = true;
                        GameProperties.Instance.triggerOptimize = false;

                    }
                    #endregion

                    #region simulate perception

                    #region calculate new?
                    // calculate new only when simulatedView is first activated
                    if (GameProperties.Instance.simulatedView == true && GameProperties.Instance.simulate == false && GameProperties.Instance.triggerSimulate)
                    {
                        GameProperties.Instance.simulate = true;
                    }
                    else
                    {
                        GameProperties.Instance.simulate = false;
                    }
                    #endregion

                    if (GameProperties.Instance.simulate)
                    {
                        if (GameProperties.Instance.simulatedView && GameProperties.Instance.triggerSimulate)
                        {
                            #region instanciate Image class
                            if (simulatedImage == null)
                            {
                                simulatedImage = new Image(GraphicsDevice, GameProperties.Instance.CVD);
                            }
                            #endregion

                            #region assign shader
                            if (simulatedImage.RGBtoLab_Shader == null)
                            {
                                simulatedImage.RGBtoLab_Shader = sim_RGBtoLab_Shader;
                                simulatedImage.LMSShader = lMSShader;
                                simulatedImage.LabToRGB_Shader = sim_LabToRGB_Shader;
                            }
                            #endregion
                            
                            #region debug: set test image to simulate perception
                            if (debugShader)
                            {
                                //simulatedImage.ImageToSimulate = rgbTestTex;
                                if (GameProperties.Instance.viewportMode.Equals(ViewportMode.Split4))
                                {
                                    simulatedImage.ImageToSimulate = dvrTexture;
                                }
                                else
                                {
                                    simulatedImage.ImageToSimulate = sceneTarget;
                                }
                            }
                            else
                            {
                                if (GameProperties.Instance.viewportMode.Equals(ViewportMode.Split4))
                                {
                                    simulatedImage.ImageToSimulate = dvrTexture;
                                    //saveScreenshot(dvrTarget, "dvrtest.png", "PNG");
                                }
                                else
                                {
                                    simulatedImage.ImageToSimulate = sceneTarget;
                                }
                            }
                            #endregion

                            // simulate perception for image
                            simulatedImage.SimulatePerception(GraphicsDevice);
                            
                            GameProperties.Instance.simulate = false;
                            GameProperties.Instance.triggerSimulate = false; //process one for every check event 
                        }


                    }
                    #endregion
                }

            }
            #endregion
        }

        private void DrawSprite(Texture2D tex)
        {
            GraphicsDevice.Clear(Color.Black);
            GraphicsDevice.BlendState = BlendState.Additive;

            spriteBatch.Begin(0, BlendState.Additive, null, null, null, null);

            spriteBatch.Draw(tex, new Rectangle(0, 0, tex.Width, tex.Height), Color.Black);

            spriteBatch.End();
        }

        //__________________________________________________________________________________________
        #endregion

        #region Draw stuff
        //-----------------------------------DRAW()-------------------------------------------------

        protected override void Draw()
        {
            //debug
            if (debugOutput)
            {
                Console.WriteLine("Engine Draw Call " + drawcount);
            }
            drawcount++;

            GraphicsDevice.SetRenderTarget(null);
            GraphicsDevice.Viewport = defaultViewport;
            GraphicsDevice.Clear(Color.Black);

            #region draw info text on start
            if (GameProperties.Instance.showInfoText)
            {
                drawInfo();
            }
            #endregion

            #region draw data

            if (GameProperties.Instance.engineState.drawData && input != null)
            {

                #region save screenshot or recolor or simulate perception: set render target
                if (GameProperties.Instance.saveView || GameProperties.Instance.recolor || GameProperties.Instance.simulate)
                {
                    GraphicsDevice.SetRenderTarget(sceneTarget);
                    if (debugOutput)
                    {
                        Console.WriteLine("GraphicsDevice Render Target set to scenetarget");
                    }
                }
                #endregion

                //GraphicsDevice.SetRenderTarget(null);
                //GraphicsDevice.SetRenderTarget(sceneTarget);

                //split screen
                if (GameProperties.Instance.viewportMode.Equals(ViewportMode.Split4))
                {
                  
                    if (GameProperties.Instance.enableCVDView)
                    {
                        
                        #region top left->3d view
                        //process only if not recolored yet and recoloredView == false
                        if(!GameProperties.Instance.recoloredView){
                        GraphicsDevice.SetRenderTarget(dvrTarget);
                        GraphicsDevice.Viewport = defaultViewport;
                        if (render > 2)
                        {
                            #region set parameters for rendering
                            GraphicsDevice.Clear(Color.Black);
                            GraphicsDevice.BlendState = BlendState.Additive;
                            GraphicsDevice.RasterizerState = RasterizerState.CullNone;
                            #endregion

                            if (GameProperties.Instance.maximumIntensityProjection == true)
                            {
                                shader.CurrentTechnique = shader.Techniques["MaximumIntensity"];
                            }
                            else
                            {
                                GraphicsDevice.BlendState = BlendState.AlphaBlend;
                                GraphicsDevice.RasterizerState = RasterizerState.CullClockwise;
                                shader.CurrentTechnique = shader.Techniques["RayCastSimple"];
                            }
                        }

                        DrawVolume(0);
                        GraphicsDevice.SetRenderTarget(null);
                        dvrTexture = dvrTarget;
                        }
                        
                        #region draw 3d sprite for the split screen

                        GraphicsDevice.Clear(Color.Black);

                        GraphicsDevice.Viewport = defaultViewport;
                        GraphicsDevice.DepthStencilState = DepthStencilState.None;
                        spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
                        spriteBatch.Draw(dvrTexture, new Rectangle(viewports[0].X, viewports[0].Y, viewports[0].Width, viewports[0].Height), Color.White);
                        spriteBatch.End();
                        GraphicsDevice.DepthStencilState = DepthStencilState.Default;

                      
                        /*GraphicsDevice.Viewport = viewports[0];
                        if (render > 2)
                        {
                            #region set parameters for rendering
                            
                            GraphicsDevice.BlendState = BlendState.Additive;
                            GraphicsDevice.RasterizerState = RasterizerState.CullNone;
                            #endregion

                            if (GameProperties.Instance.maximumIntensityProjection == true)
                            {
                                shader.CurrentTechnique = shader.Techniques["MaximumIntensity"];
                            }
                            else
                            {
                                GraphicsDevice.BlendState = BlendState.AlphaBlend;
                                GraphicsDevice.RasterizerState = RasterizerState.CullClockwise;
                                shader.CurrentTechnique = shader.Techniques["RayCastSimple"];
                            }
                        }

                        DrawVolume(0);
                        drawText(0);*/
                        #endregion

                        #endregion  
           

                        #region top right-> recolored
                        
                        if (GameProperties.Instance.recoloredView && recoloredImage.RecoloredImage != null)
                        {
                            GraphicsDevice.Viewport = defaultViewport;
                            GraphicsDevice.DepthStencilState = DepthStencilState.None;
                            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
                            spriteBatch.Draw(recoloredImage.RecoloredImage, new Rectangle(viewports[1].X, viewports[1].Y, viewports[1].Width, viewports[1].Height), Color.White);
                            spriteBatch.End();
                            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
                        
                        }
                        

                        #endregion
   
                        
                        #region bottom left-> cvd

                        GraphicsDevice.Viewport = viewports[2];
                        if (GameProperties.Instance.cvdView && GameProperties.Instance.CVDTFOptimized)
                        {
                             GraphicsDevice.SetRenderTarget(null);
                            #region draw cvd friendly volume
                            GraphicsDevice.BlendState = BlendState.AlphaBlend;
                            GraphicsDevice.RasterizerState = RasterizerState.CullClockwise;
                            try
                            {
                                foreach (ModelMesh mesh in cubeModel.Meshes)
                                {
                                    foreach (ModelMeshPart mmp in mesh.MeshParts)
                                    {
                                        mmp.Effect = CVDshader;
                                        CVDshader.SetEffectParameter_CVD(input.VolumeData,
                                            input.volumeGradients,
                                            input.stepSize * input.scaleSteps,
                                            input.iterations,
                                            input.scaleFactor,
                                            new Vector2(input.DataDimensions.X, input.DataDimensions.Y),
                                            recoloredImage.RotationToLMS,
                                            recoloredImage.RotationFromLMS);
                                        CVDshader.SetEffectParameter(input.worldMatrix * cameras[0].ViewProjection, input.worldMatrix);
                                        CVDshader.SetEffectParameter(cameras[0].CameraPosition);

                                    };
                                    initializeTextureFilter();
                                    mesh.Draw();
                                }
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine("ERROR: " + e);
                                throw;
                            }
                        }
                         #endregion
                        #endregion
                      

                        #region bottom right-> simulated

                        if (simulatedImage != null)
                        {
                            GraphicsDevice.Viewport = defaultViewport;
                            GraphicsDevice.DepthStencilState = DepthStencilState.None;
                            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
                            spriteBatch.Draw(simulatedImage.RecoloredImage, new Rectangle(viewports[3].X, viewports[3].Y, viewports[3].Width, viewports[3].Height), Color.White);
                            spriteBatch.End();
                            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
                        }

                        #endregion
                        GraphicsDevice.Viewport = viewports[0];
                        drawText(0);
                        GraphicsDevice.Viewport = viewports[1];
                        drawText(4);
                        GraphicsDevice.Viewport = viewports[2];
                        drawText(5);
                        GraphicsDevice.Viewport = viewports[3];
                        drawText(6);
                        GraphicsDevice.Viewport = defaultViewport;
                       
                      
                       
                    }
                    else //render 3d view and slices in split screen
                    {
                        if (render > 2)
                        {
                            #region set parameters for rendering
                            GraphicsDevice.Clear(Color.Black);
                            GraphicsDevice.BlendState = BlendState.Additive;
                            GraphicsDevice.RasterizerState = RasterizerState.CullNone;
                            #endregion

                            for (int current = 0; current < 4; current++)
                            {
                                GraphicsDevice.Viewport = viewports[current];

                                #region set correct shader technique
                                if (current == 1)
                                {
                                    input.slice = GameProperties.Instance.ySliceValue;
                                    shader.CurrentTechnique = shader.Techniques["RenderSlice"];
                                }
                                else if (current == 2)
                                {
                                    input.slice = GameProperties.Instance.xSliceValue;
                                    shader.CurrentTechnique = shader.Techniques["RenderSlice"];
                                }
                                else if (current == 3)
                                {
                                    input.slice = GameProperties.Instance.zSliceValue;
                                    shader.CurrentTechnique = shader.Techniques["RenderSlice"];
                                }
                                else
                                {
                                    if (GameProperties.Instance.maximumIntensityProjection == true)
                                    {
                                        shader.CurrentTechnique = shader.Techniques["MaximumIntensity"];
                                    }
                                    else
                                    {
                                        GraphicsDevice.BlendState = BlendState.AlphaBlend;
                                        GraphicsDevice.RasterizerState = RasterizerState.CullClockwise;
                                        shader.CurrentTechnique = shader.Techniques["RayCastSimple"];
                                    }
                                }
                                #endregion

                                DrawVolume(current);


                                #region draw text
                                drawText(current);
                                #endregion

                            }

                            GraphicsDevice.Viewport = defaultViewport;
                        }
                        else 
                        {
                           // swapBuffers = false;
                        }


                    }
                   
                   
                }
                 //single screens
                else if(GameProperties.Instance.viewportMode.Equals(ViewportMode.Single))
                {
                    //GraphicsDevice.SetRenderTarget(null);
                    if (debugOutput)
                    {
                        Console.WriteLine("DRAWING DATA!");
                    }
                    if (render > 2)
                    {
                        #region Draw Volume

                        #region set parameters for rendering
                        GraphicsDevice.Clear(Color.Black);
                        GraphicsDevice.BlendState = BlendState.Additive;
                        GraphicsDevice.RasterizerState = RasterizerState.CullNone;
                        //shader.SetEffectParameter(input.worldMatrix * cameras[currentView].ViewProjection, input.worldMatrix); 
                        //shader.SetEffectParameter(cameras[currentView].CameraPosition);
                        #endregion

                        #region set correct shader technique
                        if (currentView == 1)
                        {
                            input.slice = GameProperties.Instance.ySliceValue;
                            shader.CurrentTechnique = shader.Techniques["RenderSlice"];
                        }
                        else if (currentView == 2)
                        {
                            input.slice = GameProperties.Instance.xSliceValue;
                            shader.CurrentTechnique = shader.Techniques["RenderSlice"];
                        }
                        else if (currentView == 3)
                        {
                            input.slice = GameProperties.Instance.zSliceValue;
                            shader.CurrentTechnique = shader.Techniques["RenderSlice"];
                        }
                        else
                        {
                            if (GameProperties.Instance.maximumIntensityProjection == true)
                            {
                                shader.CurrentTechnique = shader.Techniques["MaximumIntensity"];
                            }
                            else
                            {
                                GraphicsDevice.BlendState = BlendState.AlphaBlend;
                                GraphicsDevice.RasterizerState = RasterizerState.CullClockwise;
                                shader.CurrentTechnique = shader.Techniques["RayCastSimple"];
                            }
                        }
                        #endregion
                        if (debugOutput)
                        {
                            Console.WriteLine(shader.CurrentTechnique.Name);

                        }

                        DrawVolume(currentView);

                        #endregion

                        #region draw text
                        if (!(GameProperties.Instance.recolor || GameProperties.Instance.simulate))
                        {
                            //draw view mode
                            drawText();
                        }
                        #endregion
                    }
                }


                #region save screenshot
                if (GameProperties.Instance.saveView)
                {
                    if (string.IsNullOrEmpty(GameProperties.Instance.saveFilename))
                    {
                        throw new ArgumentNullException(GameProperties.Instance.saveFilename);
                    }
                    GraphicsDevice.SetRenderTarget(null);
                    saveScreenshot(sceneTarget, GameProperties.Instance.saveFilename, "png");
                    GameProperties.Instance.saveView = false;
                    this.swapBuffers = false;
                    
                }
                #endregion

                GameProperties.Instance.engineState.drawData = false;

                if (debugOutput)
                {
                    Console.WriteLine("end of drawing");
                }              
            }
            #endregion
            
        }

        #region DrawVolume()
        /// <summary>
        /// Draws the Texture3D onto a cube and sets necessary EffectParameters for the shader.
        /// </summary>
        /// <param name="gameTime"></param>
        /// <param name="side"></param>
        public void DrawVolume(int side)
        {
            // draw cubeModel
            if (debugOutput)
            {
                Console.WriteLine("DrawVolume()");
            }

            

            #region draw original volume
            try
            {
                foreach (ModelMesh mesh in cubeModel.Meshes)
                {
                    foreach (ModelMeshPart mmp in mesh.MeshParts)
                    {
                        mmp.Effect = shader;
                        shader.SetEffectParameter(input.VolumeData, input.volumeGradients, input.stepSize * input.scaleSteps, input.iterations, side, input.scaleFactor, input.slice);
                        shader.SetEffectParameter(input.worldMatrix * cameras[side].ViewProjection, input.worldMatrix); 
                        shader.SetEffectParameter(cameras[side].CameraPosition);
                        if (debugOutput)
                        {
                            //Console.WriteLine("Camera pos:" + cameras[currentView].CameraPosition.X + "," + cameras[currentView].CameraPosition.Y + "," + cameras[currentView].CameraPosition.Z);
                            Console.WriteLine("drawing...side.." + side);
                        }
                    };

                     mesh.Draw();
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("ERROR: " + e);
                throw;
            }
            #endregion

        }
        #endregion

        #region save screenshot
        /// <summary>
        /// Saves the render target texture as png image.
        /// </summary>
        /// <param name="texture"></param>
        /// <param name="name"></param>
        private void saveScreenshot(Texture2D texture, String name, String ext)
        {
            if (debugOutput)
            {
                Console.WriteLine("saving " + name);
            }
            try
            {
                if (!name.ElementAt(name.Length - 4).Equals(".")) name = name + "." + ext;
                FileStream stream = File.OpenWrite(name);
                switch (ext)
                {
                    case "png":
                        texture.SaveAsPng(stream, texture.Width, texture.Height);
                        break;
                    case "jpg":
                        texture.SaveAsJpeg(stream, texture.Width, texture.Height);
                        break;
                    default:
                        if (name.ElementAt(name.Length - 4).Equals(".")) name = name.Substring(0, name.Length - 4);
                        name = name + ".png";
                        texture.SaveAsPng(stream, texture.Width, texture.Height);
                        break;
                }

            }
            catch (Exception e)
            {
                Console.WriteLine("Save Screenshot: ");
                Console.WriteLine(e);
            }

        }
        #endregion

        #region draw text
        private void drawText()
        {
            if (debugOutput)
            {
                Console.WriteLine("DRAWING TEXT!");
            }
            GraphicsDevice.DepthStencilState = DepthStencilState.None;

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);

            switch ((int)GameProperties.Instance.viewMode)
            {
                case 0:
                    spriteBatch.DrawString(spriteBatchFont, ("3D View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 1:
                    spriteBatch.DrawString(spriteBatchFont, ("Top View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 2:
                    spriteBatch.DrawString(spriteBatchFont, ("Side View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 3:
                    spriteBatch.DrawString(spriteBatchFont, ("Front View"),
                    new Vector2(50,50), Color.Silver);
                    break;
                default:
                    spriteBatch.DrawString(spriteBatchFont, ("3d View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
            }

            spriteBatch.End();
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
        }


        private void drawText(int viewMode)
        {
            if (debugOutput)
            {
                Console.WriteLine("DRAWING TEXT!");
            }
            GraphicsDevice.DepthStencilState = DepthStencilState.None;

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);

            switch (viewMode)
            {
                case 0:
                    spriteBatch.DrawString(spriteBatchFont, ("3D View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 1:
                    spriteBatch.DrawString(spriteBatchFont, ("Top View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 2:
                    spriteBatch.DrawString(spriteBatchFont, ("Side View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 3:
                    spriteBatch.DrawString(spriteBatchFont, ("Front View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 4:
                    spriteBatch.DrawString(spriteBatchFont, ("Recolored Image"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 5:
                    spriteBatch.DrawString(spriteBatchFont, ("CVD View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                case 6:
                    spriteBatch.DrawString(spriteBatchFont, ("Simulated Perception"),
                    new Vector2(50, 50), Color.Silver);
                    break;
                default:
                    spriteBatch.DrawString(spriteBatchFont, ("3d View"),
                    new Vector2(50, 50), Color.Silver);
                    break;
            }

            spriteBatch.End();
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
        }
        #endregion

        #region draw framework info
        /// <summary>
        /// Draws some framework info on application start.
        /// </summary>
        private void drawInfo()
        {
            if (debugOutput)
            {
                Console.WriteLine("DRAWING INFO!");
            }

            GraphicsDevice.Clear(Color.Black);
            GraphicsDevice.DepthStencilState = DepthStencilState.None;

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            
            //if(GameProperties.Instance.viewportMode.Equals(ViewportMode.Single))
            //{
                spriteBatch.Draw(frameworkInfo, new Rectangle(0, 0, defaultViewport.Width, defaultViewport.Height), Color.White);

            //}else if(GameProperties.Instance.viewportMode.Equals(ViewportMode.Split4))
            //{
              //  spriteBatch.Draw(frameworkInfo, new Rectangle(0, 0, viewports[0].Width, viewports[0].Height), Color.White);
            //}

            spriteBatch.DrawString(spriteBatchFont, ("Load model: File -> Open"),
                    new Vector2(30, (Height * 0.5f)), Color.Black);
            spriteBatch.DrawString(spriteBatchFont, ("Save Screenshot: File -> Save"),
                    new Vector2(30, (Height * 0.5f) + 30), Color.Black);
            spriteBatch.DrawString(spriteBatchFont, ("Load or Save TF: Functions -> Transferfunction->"),
        new Vector2(30, (Height * 0.5f) + 60), Color.Black);
            spriteBatch.DrawString(spriteBatchFont, ("CVD DVR: Extended Features -> CVD"),
        new Vector2(30, (Height * 0.5f) + 90), Color.Black);
            spriteBatch.DrawString(spriteBatchFont, ("3d View Camera: Left mouse button -> rotate, right mouse button->zoom"),
       new Vector2(30, (Height * 0.5f) + 120), Color.Black);

            spriteBatch.End();
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;

        }
        //__________________________________________________________________________________________
        #endregion

        #endregion

        #region additional Methods
        //##########################################################################################

        #region set initial control points
        void setInitialControlPoints()
        {
            //set initial control points
            GameProperties.Instance.colorControlPoints.Clear();
            GameProperties.Instance.alphaControlPoints.Clear();
            GameProperties.Instance.colorControlPoints.Add(new TransferControlPoint(0.0f, 0.0f, 0.0f, 20));//bottom left
            GameProperties.Instance.colorControlPoints.Add(new TransferControlPoint(0.0f, 0.0f, 0.0f, 250));//bottom right
            GameProperties.Instance.colorControlPoints.Add(new TransferControlPoint(1.0f, 0.0f, 0.0f, 50));
            GameProperties.Instance.colorControlPoints.Add(new TransferControlPoint(0.0f, 1.0f, 0.0f, 150));
            GameProperties.Instance.colorControlPoints.Add(new TransferControlPoint(0.0f, 0.0f, 1.0f, 200));
            GameProperties.Instance.alphaControlPoints.Add(new TransferControlPoint(0.0f, 20));//bottom left
            GameProperties.Instance.alphaControlPoints.Add(new TransferControlPoint(0.0f, 250));//bottom right
            GameProperties.Instance.alphaControlPoints.Add(new TransferControlPoint(0.8f, 50));
            GameProperties.Instance.alphaControlPoints.Add(new TransferControlPoint(1.0f, 150));
            GameProperties.Instance.alphaControlPoints.Add(new TransferControlPoint(0.8f, 200));
        }
        #endregion

        #region Initialize Texture Filter -> NEED THIS???
        /// <summary>
        /// Sets TextureFilter.
        /// </summary>
        private void initializeTextureFilter()
        {
            if (debugOutput)
            {
                Console.WriteLine("INITIALIZING TEXTURE FILTER");
            }
            /*

                device.SamplerStates[0].MipFilter = TextureFilter.Linear;
                Console.WriteLine("using Linear MipFilter");
                device.SamplerStates[0].MinFilter = TextureFilter.Linear;
                Console.WriteLine("using Linear MinFilter");
                device.SamplerStates[0].MagFilter = TextureFilter.Linear;
                Console.WriteLine("using Linear MagFilter");
             */
            try //not sure it works like this 
            {
                SamplerState stAnisotropic = new SamplerState();
                stAnisotropic.Filter = TextureFilter.Anisotropic;
                GraphicsDevice.SamplerStates[0] = stAnisotropic;
                GraphicsDevice.SamplerStates[1] = stAnisotropic;
                GraphicsDevice.SamplerStates[2] = stAnisotropic;
                GraphicsDevice.SamplerStates[3] = stAnisotropic;
                GraphicsDevice.SamplerStates[4] = stAnisotropic;
                /*SamplerState stPoint = new SamplerState();
                stPoint.Filter = TextureFilter.Point;
                device.SamplerStates[0] = stPoint;*/
            }
            catch (Exception e)
            {
                Console.WriteLine("-----------LOOK HERE:" + e);
                /*SamplerState stLinear = new SamplerState();
                stLinear.Filter = TextureFilter.Linear;
                device.SamplerStates[0] = stLinear;*/
            }
            
        }

        #endregion
        
        //__________________________________________________________________________________________

        #endregion        
  
       #endregion

        #region Control Properties Events
        protected override void OnSizeChanged(EventArgs e)
        {
            if (this.Size.Width != lastSize[0] || this.Size.Height != lastSize[1])
            {
        
                for (int i = 0; i < cameras.Length; i++)
                {
                    lastSize[0] = this.Size.Width;
                    lastSize[1] = this.Size.Height;
                    defaultViewport.Width = this.Size.Width;
                    defaultViewport.Height= this.Size.Height;
                    setViewports();
                    cameras[i].SetClientSize = lastSize;
                    
                }
                 if (debugOutput)
                {
                    Console.WriteLine("Size:" + Size);
                }

            }
            
            GameProperties.Instance.engineState.updateCamera = true;
            GameProperties.Instance.engineState.updateData = true;
            GameProperties.Instance.engineState.drawData = true;
            this.Refresh();
            base.OnSizeChanged(e);
        }

        /// <summary>
        /// Set the viewports for split screen.
        /// </summary>
        private void setViewports()
        {
            viewports = new Viewport[4];
            //top left
            viewports[0] = defaultViewport;
            viewports[0].X = 0;
            viewports[0].Y = 0;
            viewports[0].Width = defaultViewport.Width / 2;
            viewports[0].Height = defaultViewport.Height / 2;
            //top right
            viewports[1] = defaultViewport;
            viewports[1].X = defaultViewport.Width / 2; ;
            viewports[1].Y = 0;
            viewports[1].Width = defaultViewport.Width / 2;
            viewports[1].Height = defaultViewport.Height / 2;
            //bottom left
            viewports[2] = defaultViewport;
            viewports[2].X = 0;
            viewports[2].Y = defaultViewport.Height / 2;
            viewports[2].Width = defaultViewport.Width / 2;
            viewports[2].Height = defaultViewport.Height / 2;
            //bottom right
            viewports[3] = defaultViewport;
            viewports[3].X = defaultViewport.Width / 2;
            viewports[3].Y = defaultViewport.Height / 2;
            viewports[3].Width = defaultViewport.Width / 2;
            viewports[3].Height = defaultViewport.Height / 2; 
        }
        #endregion
    }
}


