﻿#region Using Statements
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
#endregion

namespace visLU
{
    /// <summary>
    /// Structure to describe the engine states
    /// </summary>
    public struct EngineState
    {
        public bool updateCamera;
        public bool loadData;
        public bool updateData;
        public bool drawData;

        public EngineState(bool _updateCamera, bool _loadData, bool _updateData, bool _drawData)
        {
            updateCamera = _updateCamera;
            loadData = _loadData;
            updateData = _updateData;
            drawData = _drawData;
        }
       
    }

    /// <summary>
    ///Enumeration to define the current view mode
    /// </summary>
    public enum ViewMode
    {
        VolView = 0,  //3d view
        TopView,     //top view -> y slice
        SideView,    //side view -> x slice
        FrontView    //front view -> z slice
    };

    /// <summary>
    ///Enumeration to define the current cvd view mode, if cvd is enabled
    /// </summary>
    public enum CVDViewMode
    {
        RefView = 0, //3d view
        RecView, //Recolored image view
        CVDView, //optimized TF
        SimView //dichromacy simulation view
    };

    /// <summary>
    /// Enumeration to set single or split screen.
    /// </summary>
    public enum ViewportMode
    {
        Single = 0,
        Split4
    };

    /// <summary>
    ///Enumeration for the different cvd types
    /// </summary>
    public enum CVDType
    {
        Deuteranopy = 0,  
        Protanopy,
        Tritanopy
    };

    /// <summary>
    /// Describes the global application properties
    /// </summary>
    public sealed class GameProperties
    {

        static GameProperties instance = new GameProperties();

        public bool shaderDebugMode = false;
        public bool debugOutput = false;

        public string CVD = "deuteranopy";
        public bool recolor = false;

       

        #region engine states
        //public bool fullscreen = false; //old version -> fullscreen option
        public EngineState engineState = new EngineState(true, false, false, false);
        #endregion

        #region xna draw window
        public int GuiWindowWidth = 1024;
        public int GuiWindowHeight = 768;
        public int XNADrawSurfaceWidth = 690;
        public int XNADrawSurfaceHeight = 700;
        public int fullscreenWidth = 1024;
        public int fullscreenHeight = 786;
       //public Viewport XNAViewport;
        public bool showInfoText = true;
        #endregion

        #region transfer function

        public float[] distinctDensityValues;
        public int[] countDensityValues;
        public List<TransferControlPoint> colorControlPoints = new List<TransferControlPoint>();
        public List<TransferControlPoint> alphaControlPoints = new List<TransferControlPoint>();
        public List<TransferControlPoint> cvdColorControlPoints = new List<TransferControlPoint>();
        #endregion

        #region gui input
        //main menu
        //initial dataset to load
        public String dataDir = "Data\\";
        public String dataFilename = "lobster.dat";
        //public String dataFilename = "skewed_head.dat";
       // public String saveFilename = "Data\\Screenshots";

        public String saveFilename = "scene.png";//for debugging
        public bool saveView = false;

        //view panel
        public ViewMode viewMode = ViewMode.VolView; //initial view: 3d view
        public CVDViewMode cvdViewMode = CVDViewMode.RefView;
        public ViewportMode viewportMode = ViewportMode.Split4;
        //slice setup
        public int xDataRange = 100;
        public int yDataRange = 100;
        public int zDataRange = 100;
        public float xSliceValue = 0.5f;
        public float ySliceValue = 0.5f;
        public float zSliceValue = 0.5f;

        //transferfunction panel
        public bool enableTransferFunction = true;
        public bool enableTransferAlphaValues = false;
        public bool showDensityValue = true;

        //cvd functionality
        //cvd functionality
        public bool CVDTFOptimized = false; //indicates if the optimization was already processed
        public CVDType cvdType = CVDType.Deuteranopy;
        public bool enableCVDView = false;
        public bool simulate = false; 
        public bool originalView = true;  //render original rgb image == ViewMode.VolView
        public bool recoloredView = false; //render recolored image
        public bool cvdView = false; //render cvd image
        public bool simulatedView = false;
        public float lambda = 0.0f; //default value by paper
        //trigger recolor, optimize and simulate
        public bool triggerRecolor = false;
        public bool triggerOptimize = false;
        public bool triggerSimulate = false;
        
        //clipping plane setup
        public bool enableClippingPlane = true; //?

        //projection setup
        public bool perspectiveProjection = true;
        public bool maximumIntensityProjection = false;

        //shader parameters
        //Blending
        public bool enableFrontToBackBlend = false;

        //Shading on/off
        public bool enableShading = true;
        public Vector3 directLightPos = new Vector3(1.0f, 4.0f, 1.0f);

        //Gradient stuff
        public int gradientSampleSize = 1;
        public float currSampleDist = 1.0f;
        public float baseSampleDist = 0.5f; //dataset sample distance
        public bool enableAvgFilter = false;
        public bool enableGaussFilter = false;

        //move camera in 3d view
        //x,y -> cursor position, z ->scroll step for zooming option
        public Vector3 position = Vector3.Zero;
        public Vector3 cameraInput = Vector3.Zero;
        public Vector3 dataOrigin = Vector3.Zero; //new Vector3(0.5f, 0.5f, 0.0f);
        #endregion
           
        //--------------------------------------------------------------
        public static GameProperties Instance
        {
            get
            {
                return instance;
            }
        }

        private static void calculateViewportDimension()
        {

        }

        static GameProperties(){}

        GameProperties(){}
    }
}
