﻿#region using statements
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using Microsoft.Xna.Framework;
#endregion

namespace visLU.Gui
{
    /// <summary>
    /// class to process input requests
    /// </summary>
    public class InputManager
    {

        bool xnaDrawSurface_MouseDownEvent = false;
        int mouseDownPosX, mouseDownPosY, mouseDownPosZ = 0;
        Vector3 mouseMovePos;

        #region properties
        public bool XnaDrawSurface_MouseDownEvent
        {
            set { xnaDrawSurface_MouseDownEvent = value; }
            get { return xnaDrawSurface_MouseDownEvent; }
            
        }

        public int MouseDownPosX
        {
            set { mouseDownPosX = value; }
        }
        public int MouseDownPosY
        {
            set { mouseDownPosY = value; }
        }
        public Vector3 MouseMovePos
        {
            set { MouseMovePos = value; }
        }
        #endregion

        /// <summary>
        /// switch the current view mode
        /// </summary>
        /// <param name="newViewMode"></param>
        public void setViewMode(ViewMode newViewMode)
        {
            GameProperties.Instance.viewMode = newViewMode;
        }

        /// <summary>
        /// return the current view mode
        /// </summary>
        /// <param name="newViewMode"></param>
        public int getViewMode()
        {
            return (int)GameProperties.Instance.viewMode;
        }


        /// <summary>
        /// switch the current view mode taking into account the cursor position
        /// </summary>
        /// <param name="posX">cursor x-Position</param>
        /// <param name="posY">cursor y-Position</param>
        public void setViewMode()
        {

             ViewMode currentViewMode = GameProperties.Instance.viewMode;
             switch (currentViewMode)
             {
                    case ViewMode.VolView:
                        GameProperties.Instance.viewMode = ViewMode.TopView;
                        break;
                    case ViewMode.TopView:
                        GameProperties.Instance.viewMode = ViewMode.SideView;
                        break;
                    case ViewMode.SideView:
                        GameProperties.Instance.viewMode = ViewMode.FrontView;
                        break;
                    case ViewMode.FrontView:
                        GameProperties.Instance.viewMode = ViewMode.VolView;
                        break;
                    default:
                        GameProperties.Instance.viewMode = ViewMode.VolView;
                        break;
                }
           
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="zoomStep"></param>
        public void setZoom(int zoomStep)
        {
            //zoom for all views???
            //zoom +/-
            if (GameProperties.Instance.viewMode == ViewMode.VolView )
            {
                // : logic in the camera class
                //int numberOfTextLinesToMove = e.Delta * SystemInformation.MouseWheelScrollLines / 120;
                //scrollStep --> e.Delta, 120->max zoom value ex. 10
                //if endvalue > 0 -> zoom + , < 0 zoom -, = 0 nothing
                GameProperties.Instance.position.Z = zoomStep;
               // mouseDownPosZ = scrollStep;
            }
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="xValue"></param>
        /// <param name="yValue"></param>
        /// <param name="zValue"></param>
        public void setSliceValue(decimal xValue, decimal yValue, decimal zValue)
        {
            GameProperties.Instance.xSliceValue = (float)xValue;
            GameProperties.Instance.ySliceValue = (float)yValue;
            GameProperties.Instance.zSliceValue = (float)zValue;

            if (!GameProperties.Instance.viewMode.Equals(ViewMode.VolView))
            {
                updateXnaEngineState(false, false, false, true);
            }
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="slice"></param>
        /// <param name="value"></param>
        public void setSliceValue(int slice, decimal value)
        {
            switch (slice)
            {
                case 1:
                    GameProperties.Instance.xSliceValue = (float) value;
                    break;
                case 2:
                    GameProperties.Instance.ySliceValue = (float) value;
                    break;
                case 3:
                    GameProperties.Instance.zSliceValue = (float) value;
                    break;
                default:
                    break;
            }

            if (!GameProperties.Instance.viewMode.Equals(ViewMode.VolView))
            {
                updateXnaEngineState(false, false, false, true);
            }
           
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="slice"></param>
        /// <returns></returns>
        public bool enableSliceGroup(int slice)
        {
            if (GameProperties.Instance.viewMode.Equals(ViewMode.VolView)) return false;

            switch (slice)
            {
                case 1: //x slice -> side view
                    if (GameProperties.Instance.viewMode.Equals(ViewMode.SideView)) return true;
                    break;
                case 2: //y slice -> top view
                    if (GameProperties.Instance.viewMode.Equals(ViewMode.TopView)) return true;
                    break;
                case 3: //z slice -> front view
                    if (GameProperties.Instance.viewMode.Equals(ViewMode.FrontView)) return true;
                    break;
                default: 
                    return false;
            }

            return false;
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="enable"></param>
        /// <param name="enableAlpha"></param>
        public void enableTransferFunction(bool enable, bool enableAlpha)
        {
            GameProperties.Instance.enableTransferFunction = enable;
            GameProperties.Instance.enableTransferAlphaValues = enableAlpha;
            updateXnaEngineState(false, false, true, true);

        }

        #region extended feature set
        /// <summary>
        ///
        /// </summary>
        /// <param name="enable"></param>
        public void enableClippingPlane(bool enable)
        {
            GameProperties.Instance.enableClippingPlane = enable;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="enable"></param>
        public void enableMaxIntensityProjection(bool enable)
        {
            GameProperties.Instance.maximumIntensityProjection = enable;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="enable"></param>
        public void enablePerspProjection(bool enable)
        {
            GameProperties.Instance.perspectiveProjection = enable;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// if false, back to front enabled
        /// </summary>
        /// <param name="enable"></param>
        public void enableFrontToBackBlending(bool enable)
        {
            GameProperties.Instance.enableFrontToBackBlend = enable;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// shading on/off
        /// </summary>
        /// <param name="enable"></param>
        public void enableShading(bool enable)
        {
            GameProperties.Instance.enableShading = enable;
            updateXnaEngineState(true, false, false, false);
        }

        #region old version fullscreen option
        /// <summary>
        /// Enables fullscreen mode.
        /// </summary>
        /// <param name="enable"></param>
        /*public void enableFullscreenMode(bool enable)
        {
            GameProperties.Instance.fullscreen = enable;
            updateXnaEngineState(true, false, true, true);
        }*/
        #endregion

        #endregion

        /// <summary>
        /// updates the current mouse position in the game properties to move the camere,
        /// but only for the 3d viewport
        /// </summary>
        /// <param name="posX">mouse x-Position</param>
        /// <param name="posY">mouse y-Position</param>
        public void updateMousePos(int posX, int posY)
        {
           if (GameProperties.Instance.viewMode == ViewMode.VolView)
            {
                GameProperties.Instance.position.X = mouseDownPosX - posX;
                GameProperties.Instance.position.Y = mouseDownPosY - posY;
                mouseDownPosX = posX;
                mouseDownPosY = posY;
            }
        }

        public void resetMousePos()
        {
            GameProperties.Instance.position.X = 0;
            GameProperties.Instance.position.Y = 0;
            GameProperties.Instance.position.Z = 0;
            mouseDownPosX = 0;
            mouseDownPosY = 0;
            mouseDownPosZ = 0;
        }

        /// <summary>
        /// updates the current mouse position in the game properties to zoom +/- the camera,
        /// but only for the 3d viewport
        /// </summary>
        /// <param name="scrollStep">how many steps the mouse was scrolled + / -</param>
        public void updateMouseScroll(int scrollStep)
        {
            int mousePos = getMousePos(mouseDownPosX, mouseDownPosY);

            //zoom +/-
            if (GameProperties.Instance.viewMode == ViewMode.VolView )
            {
                // : logic in the camera class
                //int numberOfTextLinesToMove = e.Delta * SystemInformation.MouseWheelScrollLines / 120;
                //scrollStep --> e.Delta, 120->max zoom value ex. 10
                //if endvalue > 0 -> zoom + , < 0 zoom -, = 0 nothing
                GameProperties.Instance.position.Z = scrollStep;
               // mouseDownPosZ = scrollStep;
            }
            //move slice
            else
            {
                if (GameProperties.Instance.viewMode == ViewMode.TopView)
                {
                    //GameProperties.Instance.ySliceValue = -scrollStep/10;
                }
                else if (GameProperties.Instance.viewMode == ViewMode.SideView)
                {
                    //GameProperties.Instance.xSliceValue = -scrollStep/10;
                }
                else if (GameProperties.Instance.viewMode == ViewMode.FrontView)
                {
                    //GameProperties.Instance.zSliceValue = -scrollStep/10;
                }

            }
 
        }
        
        /// <summary>
        /// Checks if the cursor is inside the viewing window and returns its location
        /// </summary>
        /// <param name="posX">cursor x-Position</param>
        /// <param name="posY">cursor y-Position</param>
        /// <returns> Returns an int value:
        /// -1 => cursor outside the viewport(s) so request ignored,
        ///  0 => cursor in top left (area of the) viewport,
        ///  1 => cursor in top right (area of the) viewport,
        ///  2 => cursor in bottom left (area of the) viewport,
        ///  3 => cursor in bottom right (area of the) viewport
        /// </returns>
        private int getMousePos(int posX, int posY)
        {
            /*if ((posX >= GameProperties.Instance.XNASingleViewport.X && posY >= GameProperties.Instance.XNASingleViewport.Y)
                && (posX <= GameProperties.Instance.XNASingleViewport.X + GameProperties.Instance.XNASingleViewport.Width && posY <= GameProperties.Instance.XNASingleViewport.Y + GameProperties.Instance.XNASingleViewport.Height))
            {

                for (int i = 0; i < 4; i++)
                {
                    if ((posX >= GameProperties.Instance.XNAViewports[i].X && posY >= GameProperties.Instance.XNAViewports[i].Y)
                && (posX <= GameProperties.Instance.XNAViewports[i].X + GameProperties.Instance.XNAViewports[i].Width && posY <= GameProperties.Instance.XNAViewports[i].Y + GameProperties.Instance.XNAViewports[i].Height))
                        return i;
                }
                return 1;

            }*/

            //mouse outside the viewports
            return -1;
        }

        public void loadTransferFunction(String fileName)
        {
            GameProperties.Instance.colorControlPoints.Clear();
            GameProperties.Instance.alphaControlPoints.Clear();

            XmlTextReader xmlReader = new XmlTextReader(fileName);
            xmlReader.MoveToElement();

            float r = 0.0f;
            float g = 0.0f;
            float b = 0.0f;
            float a = 0.0f; 
            int isoValue = 0;

            while (xmlReader.Read())
            {
                if (xmlReader.NodeType == XmlNodeType.Element && xmlReader.Name.Equals("ControlPoint"))
                {
                    while (!(xmlReader.NodeType == XmlNodeType.EndElement && xmlReader.Name.Equals("ControlPoint")))
                    {
                        xmlReader.Read();
                        
                        if (xmlReader.NodeType.Equals(XmlNodeType.Element))
                        {
                            String elementName = xmlReader.Name;
                            xmlReader.Read();
                            if (xmlReader.NodeType.Equals(XmlNodeType.Text))
                            {
                                switch (elementName)
                                {
                                    case "ColorR":
                                        r = float.Parse(xmlReader.Value);
                                        break;

                                    case "ColorG":
                                        g = float.Parse(xmlReader.Value);
                                        break;

                                    case "ColorB":
                                        b = float.Parse(xmlReader.Value);
                                        break;

                                    case "ColorA":
                                        a = float.Parse(xmlReader.Value);
                                        break;

                                    case "IsoValue":
                                        isoValue = int.Parse(xmlReader.Value);
                                        break;
                                }//end switch
                            }//end if
                        }
                    }//end while

                   //save new control point
                   GameProperties.Instance.colorControlPoints.Add(new TransferControlPoint(r, g, b, isoValue));
                   GameProperties.Instance.alphaControlPoints.Add(new TransferControlPoint(a, isoValue));
                }   
            }
        }

        /// <summary>
        /// saves all control points that are set 
        /// </summary>
        public void saveTransferFunction(string fileName)
        {
            XmlTextWriter xmlWriter = new XmlTextWriter(fileName, null);

            xmlWriter.WriteStartDocument();
            xmlWriter.WriteStartElement("TransferFunction");

            for (int i = 0; i < GameProperties.Instance.colorControlPoints.Count(); i++)
            {
                xmlWriter.WriteStartElement("ControlPoint");

                xmlWriter.WriteStartElement("ColorR");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].color.X.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("ColorG");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].color.Y.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("ColorB");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].color.Z.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("ColorA");
                xmlWriter.WriteString(GameProperties.Instance.alphaControlPoints[i].color.W.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("IsoValue");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].isoValue.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteEndElement();
            }

            xmlWriter.WriteEndElement();
            xmlWriter.WriteEndDocument();
            xmlWriter.Close();
        }

        #region update game and engine state 
        public void updateXnaEngineState(bool updateCam, bool load, bool update, bool draw) 
        {
            GameProperties.Instance.engineState.updateCamera = updateCam;
            GameProperties.Instance.engineState.loadData = load;
            GameProperties.Instance.engineState.updateData = update;
            GameProperties.Instance.engineState.drawData = draw;
        }
        #endregion
    }
}
