﻿#region using statemens
using System;
using System.IO;
using System.Reflection;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Xml;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using visLU.Gui;
#endregion

namespace visLU.Gui
{
    using Point = System.Drawing.Point;

    /// <summary>
    ///  Main Window Class.
    /// </summary>
    public partial class MainWindow : Form
    {
        #region Tranfer function variables
        bool tfPanelMouseDown = false;
        Point tfPanelMouseDownPos = new Point();
        #endregion

        #region XNA Viewport variables
        bool xnaDrawSurface_MouseDownEvent = false;
        int mouseDownPosX, mouseDownPosY, mouseDownPosZ = 0;
        Vector3 mouseMovePos;
        #endregion

        #region other variables
        string assemblyLocation = Assembly.GetExecutingAssembly().Location;

        SlicingForm slicingDialog;
        bool sliceFormOpen;
        #endregion

        #region Properties
        public bool XnaDrawSurface_MouseDownEvent
        {
            set { xnaDrawSurface_MouseDownEvent = value; }
            get { return xnaDrawSurface_MouseDownEvent; }

        }

        public int MouseDownPosX
        {
            set { mouseDownPosX = value; }
        }
        public int MouseDownPosY
        {
            set { mouseDownPosY = value; }
        }
        public Vector3 MouseMovePos
        {
            set { MouseMovePos = value; }
        }

        public bool SliceFormOpen
        {
            set { sliceFormOpen = value; }
            get { return sliceFormOpen; }
        }

        #endregion

        /// <summary>
        /// 
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();
            SetStyle(ControlStyles.OptimizedDoubleBuffer |
                        ControlStyles.UserPaint |
                        ControlStyles.AllPaintingInWmPaint, true);

            #region init child controls
            slicingDialog = null;
            sliceFormOpen = false;
            this.transferfunctionCheckBox.Checked = GameProperties.Instance.enableTransferFunction;
            this.transferfunctionCheckBox1.Checked = GameProperties.Instance.enableTransferAlphaValues;
            this.transferfunctionCheckBox2.Checked = GameProperties.Instance.showDensityValue;
            this.enableCVDCheckBox.Checked = GameProperties.Instance.enableCVDView;
            this.cvdViewCheckBox1.Checked = GameProperties.Instance.originalView;
            this.cvdViewCheckBox2.Checked = GameProperties.Instance.recoloredView;
            this.cvdViewCheckBox3.Checked = GameProperties.Instance.cvdView;
            this.cvdViewCheckBox4.Checked = GameProperties.Instance.simulate;
            //this.cvdViewToolStripMenuItem.Checked = GameProperties.Instance.cvdView;
            //this.recoloredViewToolStripMenuItem.Checked = GameProperties.Instance.recoloredView;
            switch (GameProperties.Instance.cvdType)
            {
                case CVDType.Deuteranopy:
                    this.cvdRadioButton1.Checked = true;
                    this.cvdRadioButton2.Checked = false;
                    this.cvdRadioButton3.Checked = false;
                    break;
                case CVDType.Protanopy:
                    this.cvdRadioButton1.Checked = false;
                    this.cvdRadioButton2.Checked = true;
                    this.cvdRadioButton3.Checked = false;
                    break;
                case CVDType.Tritanopy:
                    this.cvdRadioButton1.Checked = false;
                    this.cvdRadioButton2.Checked = false;
                    this.cvdRadioButton3.Checked = true;
                    break;
                default:
                    this.cvdRadioButton1.Checked = true;
                    this.cvdRadioButton2.Checked = false;
                    this.cvdRadioButton3.Checked = false;
                    break;
            }

            if (GameProperties.Instance.enableCVDView)
            {
                this.cvdViewCheckBox1.Enabled = true;
                this.cvdViewCheckBox2.Enabled = true;
                this.cvdViewCheckBox3.Enabled = true;
                //this.cvdViewToolStripMenuItem.Enabled = true;
                //this.recoloredViewToolStripMenuItem.Enabled = true;
                this.cvdRadioButton1.Enabled = true;
                this.cvdRadioButton2.Enabled = true;
                this.cvdRadioButton3.Enabled = true;
            }
            else 
            {
                this.cvdViewCheckBox1.Enabled = false;
                this.cvdViewCheckBox2.Enabled = false;
                this.cvdViewCheckBox3.Enabled = false;
                //this.cvdViewToolStripMenuItem.Enabled = false;
                //this.recoloredViewToolStripMenuItem.Enabled = false;
                this.cvdRadioButton1.Enabled = false;
                this.cvdRadioButton2.Enabled = false;
                this.cvdRadioButton3.Enabled = false;
            }

            switch (GameProperties.Instance.viewMode)
            {
                case ViewMode.VolView:
                    volViewToolStripMenuItem.Checked = true;
                    break;
                case ViewMode.TopView:
                    topViewToolStripMenuItem.Checked = true;
                    break;
                case ViewMode.SideView:
                    sideViewToolStripMenuItem.Checked = true;
                    break;
                case ViewMode.FrontView:
                    frontViewToolStripMenuItem.Checked = true;
                    break;
                default:
                    volViewToolStripMenuItem.Checked = true;
                    break;
            }

            this.PerspProjRadio.Checked = GameProperties.Instance.perspectiveProjection;
            this.OrthProjRadio.Checked = !GameProperties.Instance.perspectiveProjection;
            this.BlendRadioButt1.Checked = GameProperties.Instance.enableFrontToBackBlend;
            this.BlendRadioButt2.Checked = !GameProperties.Instance.enableFrontToBackBlend;
            this.ShadingCheckBox.Checked = GameProperties.Instance.enableShading;

            #endregion
        }

        #region update controls
        /// <summary>
        /// Updates the values in the slicing panel.
        /// </summary>
        /// <param name="xRange"></param>
        /// <param name="yRange"></param>
        /// <param name="ZRange"></param>
        public void updateSliceControls(int xRange, int yRange, int zRange)
        {
            //x Slice / side view
            Decimal xIncrement = Decimal.Divide(XSliceNumericUpDown.Maximum, new Decimal(xRange));
            //string xString = xIncrement.ToString();
            //int xDecimalPlaces = xString.Substring(xString.IndexOf(",") + 1).Length;
            int xDecimalPlaces = 8;
            XSliceNumericUpDown.Increment = xIncrement;
            XSliceNumericUpDown.DecimalPlaces = xDecimalPlaces;
            XSliceNumericUpDown.Value = Decimal.Divide(XSliceNumericUpDown.Maximum, new Decimal(2));
            XSlider.Maximum = xRange;
            XSlider.Value = xRange / 2;

            //y Slice / top view
            Decimal yIncrement = Decimal.Divide(YSliceNumericUpDown.Maximum, new Decimal(yRange));
            //string yString = yIncrement.ToString();
            //int yDecimalPlaces = yString.Substring(yString.IndexOf(",") + 1).Length;
            int yDecimalPlaces = 8;
            YSliceNumericUpDown.Increment = yIncrement;
            YSliceNumericUpDown.DecimalPlaces = yDecimalPlaces;
            YSliceNumericUpDown.Value = Decimal.Divide(YSliceNumericUpDown.Maximum, new Decimal(2));
            YSlider.Maximum = yRange;
            YSlider.Value = yRange / 2;

            //z Slice / front
            Decimal zIncrement = Decimal.Divide(ZSliceNumericUpDown.Maximum, new Decimal(zRange));
            //string zString = zIncrement.ToString();
            //int zDecimalPlaces = zString.Substring(zString.IndexOf(",") + 1).Length;
            int zDecimalPlaces = 8;
            ZSliceNumericUpDown.Increment = zIncrement;
            ZSliceNumericUpDown.DecimalPlaces = zDecimalPlaces;
            ZSliceNumericUpDown.Value = Decimal.Divide(ZSliceNumericUpDown.Maximum, new Decimal(2));
            ZSlider.Maximum = zRange;
            ZSlider.Value = zRange / 2;

            XSliceGroup.Refresh();
            YSliceGroup.Refresh();
            ZSliceGroup.Refresh();
        }

        /// <summary>
        /// Initializes the slider in slicing panel with the values from the slicing form.
        /// </summary>
        /// <param name="slice"></param>
        /// <param name="increment"></param>
        /// <param name="decimalPlaces"></param>
        /// <param name="value"></param>
        public void setSlicingGroups(int slice, Decimal increment, int decimalPlaces, Decimal value)
        {
            switch (slice)
            {
                case 1:
                    XSliceNumericUpDown.Increment = increment;
                    XSliceNumericUpDown.DecimalPlaces = decimalPlaces;
                    XSliceNumericUpDown.Value = value;
                    break;
                case 2:
                    YSliceNumericUpDown.Increment = increment;
                    YSliceNumericUpDown.DecimalPlaces = decimalPlaces;
                    YSliceNumericUpDown.Value = value;
                    break;
                case 3:
                    ZSliceNumericUpDown.Increment = increment;
                    ZSliceNumericUpDown.DecimalPlaces = decimalPlaces;
                    ZSliceNumericUpDown.Value = value;
                    break;
                default:
                    break;

            }

        }

        /// <summary>
        /// Initializes the slider in slicing panel with the values from the slicing form.
        /// </summary>
        /// <param name="slice"></param>
        /// <param name="maximum"></param>
        /// <param name="value"></param>
        public void setSlicingGroups(int slice, int maximum, int value)
        {
            switch (slice)
            {
                case 1:
                    XSlider.Maximum = maximum;
                    XSlider.Value = value;
                    break;
                case 2:
                    YSlider.Maximum = maximum;
                    YSlider.Value = value;
                    break;
                case 3:
                    ZSlider.Maximum = maximum;
                    ZSlider.Value = value;
                    break;
                default:
                    break;

            }

        }

        public void updateSliceGroups()
        {
            XSliceGroup.Refresh();
            YSliceGroup.Refresh();
            ZSliceGroup.Refresh();
        }
        #endregion

        #region event handler

        #region main menu handler

        #region file menu
        /// <summary>
        /// Handle for File -> Open;
        /// Initialize the load process for the selected dataset.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleOpenFileMenu(object sender, EventArgs e)
        {
            // Default to the directory which contains our content files.
            //string assemblyLocation = Assembly.GetExecutingAssembly().Location;
            string relativePath = Path.Combine(assemblyLocation, "../../../../Data");
            string contentPath = Path.GetFullPath(relativePath);

            this.openFileDialog.InitialDirectory = contentPath;

            this.openFileDialog.Title = "Load Data Set";

            this.openFileDialog.Filter = "Data Files (*.dat)|*.dat";

            if (this.openFileDialog.ShowDialog() == DialogResult.OK)
            {
                #region set engine control
                GameProperties.Instance.dataFilename = openFileDialog.FileName.Substring(openFileDialog.FileName.LastIndexOf("\\") + 1);
                updateXnaEngineState(true, true, true, true);
                #endregion

                #region set tf control
                TransferFunctionPanel.UpdateIsoValues = true;
                TransferFunctionPanel.DensityFunction = (GameProperties.Instance.showDensityValue && this.transferfunctionCheckBox2.Checked);
                TransferFunctionPanel.GetControlPoints = true;
                TransferFunctionPanel.Refresh();
                #endregion
            }
        }

        /// <summary>
        ///  Handles the save sceenshot option. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleSaveFileMenu(object sender, EventArgs e)
        {
            string relativePath = Path.Combine(assemblyLocation, "../../../../Data/Screenshots");
            string contentPath = Path.GetFullPath(relativePath);

            this.saveFileDialog.InitialDirectory = contentPath;

            this.saveFileDialog.Title = "Save Image as";

            this.saveFileDialog.Filter = "PNG Files (*.png)|*.png";

            if (this.saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                #region update engine state
                GameProperties.Instance.saveFilename = saveFileDialog.FileName;
                GameProperties.Instance.saveView = true;
                GameProperties.Instance.engineState.drawData = true;
                this.XNADrawSurface.Refresh();
                #endregion
            }
        }

        /// <summary>
        /// Exits the application.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleExitMenu(object sender, EventArgs e)
        {
            System.Windows.Forms.Application.Exit();

        }

        /// <summary>
        /// Handles some stuff before exit the application.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            //todo?
        }
        #endregion

        #region view menu
        /// <summary>
        /// Handles the View menu.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleViewMenuOnClick(object sender, EventArgs e)
        {
           if(sender == this.SingleToolStripMenuItem)
           {
               SingleToolStripMenuItem.Checked = true;
               SplitToolStripMenuItem.Checked = false;
               GameProperties.Instance.viewportMode = ViewportMode.Single;

               volViewToolStripMenuItem.Enabled = true;
               topViewToolStripMenuItem.Enabled = true;
               sideViewToolStripMenuItem.Enabled = true;
               frontViewToolStripMenuItem.Enabled = true;

           }
           else if (sender == this.SplitToolStripMenuItem)
           {
               SingleToolStripMenuItem.Checked = false;
               SplitToolStripMenuItem.Checked = true;
               GameProperties.Instance.viewportMode = ViewportMode.Split4;

               volViewToolStripMenuItem.Enabled = false;
               topViewToolStripMenuItem.Enabled = false;
               sideViewToolStripMenuItem.Enabled = false;
               frontViewToolStripMenuItem.Enabled = false;
           }

           else if (sender == this.volViewToolStripMenuItem)
           {
               volViewToolStripMenuItem.Checked = true;
               topViewToolStripMenuItem.Checked = false;
               sideViewToolStripMenuItem.Checked = false;
               frontViewToolStripMenuItem.Checked = false;
               GameProperties.Instance.viewMode = ViewMode.VolView;
               if (GameProperties.Instance.enableCVDView)
               {
                   //cvdViewCheckBox1.Checked = volViewToolStripMenuItem.Checked;
                   //recoloredViewToolStripMenuItem.Checked = GameProperties.Instance.recoloredView;
                   //cvdViewToolStripMenuItem.Checked = GameProperties.Instance.cvdView;
                   GameProperties.Instance.originalView = volViewToolStripMenuItem.Checked;
               }
           }
           else if (sender == this.topViewToolStripMenuItem)
           {
               volViewToolStripMenuItem.Checked = false;
               topViewToolStripMenuItem.Checked = true;
               sideViewToolStripMenuItem.Checked = false;
               frontViewToolStripMenuItem.Checked = false;
               recoloredViewToolStripMenuItem.Checked = false;
               //cvdViewToolStripMenuItem.Checked = false;
               GameProperties.Instance.viewMode = ViewMode.TopView;
           }
           else if (sender == this.sideViewToolStripMenuItem)
           {
               volViewToolStripMenuItem.Checked = false;
               topViewToolStripMenuItem.Checked = false;
               sideViewToolStripMenuItem.Checked = true;
               frontViewToolStripMenuItem.Checked = false;
               recoloredViewToolStripMenuItem.Checked = false;
               //cvdViewToolStripMenuItem.Checked = false;
               GameProperties.Instance.viewMode = ViewMode.SideView;
           }
           else if (sender == this.frontViewToolStripMenuItem)
           {
               volViewToolStripMenuItem.Checked = false;
               topViewToolStripMenuItem.Checked = false;
               sideViewToolStripMenuItem.Checked = false;
               frontViewToolStripMenuItem.Checked = true;
               recoloredViewToolStripMenuItem.Checked = false;
              // cvdViewToolStripMenuItem.Checked = false;
               GameProperties.Instance.viewMode = ViewMode.FrontView;
           }
           else if (sender == this.recoloredViewToolStripMenuItem)
           {
               //volViewToolStripMenuItem.Checked = false;
               topViewToolStripMenuItem.Checked = false;
               sideViewToolStripMenuItem.Checked = false;
               frontViewToolStripMenuItem.Checked = false;
               recoloredViewToolStripMenuItem.Checked = !recoloredViewToolStripMenuItem.Checked;
               //cvdViewToolStripMenuItem.Checked = false;
               GameProperties.Instance.recoloredView = recoloredViewToolStripMenuItem.Checked;
               if (GameProperties.Instance.enableCVDView)
               {
                   cvdViewCheckBox2.Checked = recoloredViewToolStripMenuItem.Checked;
               }
           }
           else if (sender == this.cvdViewToolStripMenuItem)
           {
               /*//volViewToolStripMenuItem.Checked = false;
               topViewToolStripMenuItem.Checked = false;
               sideViewToolStripMenuItem.Checked = false;
               frontViewToolStripMenuItem.Checked = false;
               //recoloredViewToolStripMenuItem.Checked = false;
               cvdViewToolStripMenuItem.Checked = !cvdViewToolStripMenuItem.Checked;
               GameProperties.Instance.cvdView = cvdViewToolStripMenuItem.Checked;
               if (GameProperties.Instance.enableCVDView)
               {
                   cvdViewCheckBox3.Checked = cvdViewToolStripMenuItem.Checked;
               }*/
           }

           // updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Handles the zoom options in the view menu. TODO
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleZoomMenuOnCLick(object sender, EventArgs e)
        {

            /*if (sender == this.zoomInToolStripMenuItem) inputManager.setZoom(1);
            if (sender == this.zoomOutToolStripMenuItem) inputManager.setZoom(-1);
            //inputManager.setZoom(e.Delta / SystemInformation.MouseWheelScrollDelta);
            //inputManager.setZoom(e.Delta);
            inputManager.updateXnaEngineState(true, false, true, true);*/
        }
        #endregion

        #region functions menu
        //slicing
        /// <summary>
        /// Opens new dialog for the slicing panel.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleSlicingMenu(object sender, EventArgs e)
        {
            slicingDialog = new SlicingForm();
            sliceFormOpen = true;
            slicingDialog.MainForm = this;
            slicingDialog.setSlicingGroups(1, XSliceNumericUpDown.Increment, XSliceNumericUpDown.DecimalPlaces, XSliceNumericUpDown.Value);
            slicingDialog.setSlicingGroups(2, YSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, YSliceNumericUpDown.Value);
            slicingDialog.setSlicingGroups(3, ZSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, ZSliceNumericUpDown.Value);
            slicingDialog.setSlicingGroups(1, XSlider.Maximum, XSlider.Value);
            slicingDialog.setSlicingGroups(2, YSlider.Maximum, YSlider.Value);
            slicingDialog.setSlicingGroups(3, ZSlider.Maximum, ZSlider.Value);
            slicingDialog.Show();
        }



        //projection functions
        /// <summary>
        /// Enables/Disables Maximum Intensity Projection
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleMaxIntensityProjMenu(object sender, EventArgs e)
        {
            if (maximumIntensityProjectionToolStripMenuItem.Checked)
            {
                maximumIntensityProjectionToolStripMenuItem.Checked = false;
                MaxIntensityProjCheckBox.Checked = false;
            }
            else
            {
                maximumIntensityProjectionToolStripMenuItem.Checked = true;
                MaxIntensityProjCheckBox.Checked = true;
            }

            GameProperties.Instance.maximumIntensityProjection = MaxIntensityProjCheckBox.Checked;
            updateXnaEngineState(true, false, true, true);

        }
        /// <summary>
        /// Switches between perspective and orthogonal projections.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handlePerspProjMenu(object sender, EventArgs e)
        {
            if (perspectiveProjectionToolStripMenuItem.Checked)
            {
                perspectiveProjectionToolStripMenuItem.Checked = false;
                orthogonalProjectionToolStripMenuItem.Checked = true;
                PerspProjRadio.Checked = false;
                OrthProjRadio.Checked = true;
            }
            else
            {
                perspectiveProjectionToolStripMenuItem.Checked = true;
                orthogonalProjectionToolStripMenuItem.Checked = false;
                PerspProjRadio.Checked = true;
                OrthProjRadio.Checked = false;
            }

            GameProperties.Instance.perspectiveProjection = PerspProjRadio.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Switches between perspective and orthogonal projections.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleOrthProjMenu(object sender, EventArgs e)
        {
            if (orthogonalProjectionToolStripMenuItem.Checked)
            {
                perspectiveProjectionToolStripMenuItem.Checked = true;
                orthogonalProjectionToolStripMenuItem.Checked = false;
                PerspProjRadio.Checked = true;
                OrthProjRadio.Checked = false;
            }
            else
            {
                perspectiveProjectionToolStripMenuItem.Checked = false;
                orthogonalProjectionToolStripMenuItem.Checked = true;
                PerspProjRadio.Checked = false;
                OrthProjRadio.Checked = true;
            }
            GameProperties.Instance.perspectiveProjection = PerspProjRadio.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Switschen between Front to Back and Back to Front Blending
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleBlendModeMenu(object sender, EventArgs e)
        {
            if (sender == frontToBackBlendingToolStripMenuItem)
            {
                if (frontToBackBlendingToolStripMenuItem.Checked)
                {
                    frontToBackBlendingToolStripMenuItem.Checked = false;
                    backToFrontBlendingToolStripMenuItem.Checked = true;
                    BlendRadioButt1.Checked = false;
                    BlendRadioButt2.Checked = true;
                }
                else
                {
                    frontToBackBlendingToolStripMenuItem.Checked = true;
                    backToFrontBlendingToolStripMenuItem.Checked = false;
                    BlendRadioButt1.Checked = true;
                    BlendRadioButt2.Checked = false;
                }

            }
            else if (sender == backToFrontBlendingToolStripMenuItem)
            {
                if (backToFrontBlendingToolStripMenuItem.Checked)
                {
                    frontToBackBlendingToolStripMenuItem.Checked = true;
                    backToFrontBlendingToolStripMenuItem.Checked = false;
                    BlendRadioButt1.Checked = true;
                    BlendRadioButt2.Checked = false;
                }
                else
                {
                    frontToBackBlendingToolStripMenuItem.Checked = false;
                    backToFrontBlendingToolStripMenuItem.Checked = true;
                    BlendRadioButt1.Checked = false;
                    BlendRadioButt2.Checked = true;
                }

            }

            GameProperties.Instance.enableFrontToBackBlend = frontToBackBlendingToolStripMenuItem.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Enable/Disable the shading mode
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleShadingMenu(object sender, EventArgs e)
        {
            if (shadingToolStripMenuItem.Checked)
            {
                shadingToolStripMenuItem.Checked = false;
            }
            else
            {
                shadingToolStripMenuItem.Checked = true;
            }
            ShadingCheckBox.Checked = shadingToolStripMenuItem.Checked;
            GameProperties.Instance.enableShading = ShadingCheckBox.Checked;
            updateXnaEngineState(true, false, true, true);
            
        }

        //transfer function
        /// <summary>
        /// Enables/Disables the transferfunction
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleEnableTransferFunctionMenu(object sender, EventArgs e)
        {
            if (enableTransferfunctionToolStripMenuItem.Checked)
            {
                enableTransferfunctionToolStripMenuItem.Checked = false;
                transferfunctionCheckBox1.Checked = false;
            }
            else enableTransferfunctionToolStripMenuItem.Checked = true;
            transferfunctionCheckBox.Checked = enableTransferfunctionToolStripMenuItem.Checked;

            #region update engine state
            GameProperties.Instance.enableTransferFunction = enableTransferfunctionToolStripMenuItem.Checked;
            GameProperties.Instance.enableTransferAlphaValues = transferfunctionCheckBox1.Checked;
            updateXnaEngineState(false, false, true, true);
            #endregion

            #region update tf control state
            TransferFunctionPanel.EnableTransferFunction = enableTransferfunctionToolStripMenuItem.Checked;
            TransferFunctionPanel.Refresh();
            #endregion
        }

        /// <summary>
        /// Loads transferfunction definition from a xml file.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleLoadTransferFunctionMenu(object sender, EventArgs e)
        {
            string relativePath = Path.Combine(assemblyLocation, "../../../../Data/Transferfunction");
            string contentPath = Path.GetFullPath(relativePath);

            this.openFileDialog.InitialDirectory = contentPath;

            this.openFileDialog.Title = "Load Transferfunction";

            this.openFileDialog.Filter = "XML Files (*.xml)|*.xml";

            if (this.openFileDialog.ShowDialog() == DialogResult.OK)
            {
                loadTransferFunction(openFileDialog.FileName);

                #region update engine state
                updateXnaEngineState(false, false, true, true);
                #endregion

                #region update transferfunction state
                TransferFunctionPanel.GetControlPoints = true;
                TransferFunctionPanel.Refresh();
                #endregion

            }
        }

        /// <summary>
        /// Saves a transferfunction definition as a xml file.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleSaveTransferFunctionMenu(object sender, EventArgs e)
        {
            string relativePath = Path.Combine(assemblyLocation, "../../../../Data/Transferfunction");
            string contentPath = Path.GetFullPath(relativePath);

            this.saveFileDialog.InitialDirectory = contentPath;

            this.saveFileDialog.Title = "Save Transferfunction as";

            this.saveFileDialog.Filter = "XML Files (*.xml)|*.xml";

            if (this.saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                saveTransferFunction(saveFileDialog.FileName);
            }
        }


        /// <summary>
        ///  TODO: ???? DO WE NEED THIS? -> DELETE???
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleEditTransferFunctionMenu(object sender, EventArgs e)
        {
            TransferfunctionForm transferFunctionDialog = new TransferfunctionForm();
            transferFunctionDialog.Show();
        }

        #endregion

        #region help menu
        // TODO????
        #endregion

        #endregion

        #region settings panel (right tab panel)

        #region slices
        /// <summary>
        /// Handle for the sclice panels.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleSliceNumericUpDown(object sender, EventArgs e)
        {
            if (sender == XSliceNumericUpDown)
            {
                XSlider.Value = Decimal.ToInt32(Decimal.Multiply(XSliceNumericUpDown.Value, new Decimal(XSlider.Maximum)));
                GameProperties.Instance.xSliceValue = (float)XSliceNumericUpDown.Value;

            }
            else if (sender == YSliceNumericUpDown)
            {
                YSlider.Value = Decimal.ToInt32(Decimal.Multiply(YSliceNumericUpDown.Value, new Decimal(YSlider.Maximum)));
                GameProperties.Instance.ySliceValue = (float)YSliceNumericUpDown.Value;

            }
            else if (sender == ZSliceNumericUpDown)
            {
                ZSlider.Value = Decimal.ToInt32(Decimal.Multiply(ZSliceNumericUpDown.Value, new Decimal(ZSlider.Maximum)));
                GameProperties.Instance.zSliceValue = (float)ZSliceNumericUpDown.Value;

            }
            else { return; }

            if (sliceFormOpen)
            {
                slicingDialog.setSlicingGroups(1, XSliceNumericUpDown.Increment, XSliceNumericUpDown.DecimalPlaces, XSliceNumericUpDown.Value);
                slicingDialog.setSlicingGroups(2, YSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, YSliceNumericUpDown.Value);
                slicingDialog.setSlicingGroups(3, ZSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, ZSliceNumericUpDown.Value);
                slicingDialog.setSlicingGroups(1, XSlider.Maximum, XSlider.Value);
                slicingDialog.setSlicingGroups(2, YSlider.Maximum, YSlider.Value);
                slicingDialog.setSlicingGroups(3, ZSlider.Maximum, ZSlider.Value);
                slicingDialog.Refresh();
            }

            if (!GameProperties.Instance.viewMode.Equals(ViewMode.VolView))
            {
                updateXnaEngineState(false, false, false, true);
            }
        }

        /// <summary>
        /// Handle for the sclice panels.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleScrollSlider(object sender, EventArgs e)
        {
            if (sender == XSlider)
            {
                Decimal newValue = Decimal.Divide(new Decimal(XSlider.Value), new Decimal(XSlider.Maximum));
                XSliceNumericUpDown.Value = newValue;
                GameProperties.Instance.xSliceValue = (float)newValue;
            }
            else if (sender == YSlider)
            {
                Decimal newValue = Decimal.Divide(new Decimal(YSlider.Value), new Decimal(YSlider.Maximum));
                YSliceNumericUpDown.Value = newValue;
                GameProperties.Instance.ySliceValue = (float)newValue;

            }
            else if (sender == ZSlider)
            {
                Decimal newValue = Decimal.Divide(new Decimal(ZSlider.Value), new Decimal(ZSlider.Maximum));
                ZSliceNumericUpDown.Value = newValue;
                GameProperties.Instance.zSliceValue = (float)newValue;

            }
            else { return; }

            if (sliceFormOpen)
            {
                slicingDialog.setSlicingGroups(1, XSliceNumericUpDown.Increment, XSliceNumericUpDown.DecimalPlaces, XSliceNumericUpDown.Value);
                slicingDialog.setSlicingGroups(2, YSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, YSliceNumericUpDown.Value);
                slicingDialog.setSlicingGroups(3, ZSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, ZSliceNumericUpDown.Value);
                slicingDialog.setSlicingGroups(1, XSlider.Maximum, XSlider.Value);
                slicingDialog.setSlicingGroups(2, YSlider.Maximum, YSlider.Value);
                slicingDialog.setSlicingGroups(3, ZSlider.Maximum, ZSlider.Value);
                slicingDialog.Refresh();
            }

            if (!GameProperties.Instance.viewMode.Equals(ViewMode.VolView))
            {
                updateXnaEngineState(false, false, false, true);
            }

        }
        #endregion

        #region  transfer function panel
        /// <summary>
        /// Handle to enable transferfunction checkbox (transferfunction panel).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleTransferFunctionCheckbox(object sender, EventArgs e)
        {
            //set gui
            if (!transferfunctionCheckBox.Checked) transferfunctionCheckBox1.Checked = false;
            enableTransferfunctionToolStripMenuItem.Checked = transferfunctionCheckBox.Checked;

            #region set engine control
            GameProperties.Instance.enableTransferFunction = transferfunctionCheckBox.Checked;
            GameProperties.Instance.enableTransferAlphaValues = transferfunctionCheckBox1.Checked;
            updateXnaEngineState(false, false, true, true);
            #endregion

            #region set tf control
            TransferFunctionPanel.EnableTransferFunction = transferfunctionCheckBox.Checked;
            TransferFunctionPanel.Refresh();
            #endregion
        }

        /// <summary>
        ///  Handle to enable alpha values checkbox (transferfunction panel).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleTransferFunctionCheckbox1(object sender, EventArgs e)
        {
            if (transferfunctionCheckBox1.Checked && !transferfunctionCheckBox.Checked) transferfunctionCheckBox1.Checked = false;

            #region set engine control
            if (GameProperties.Instance.enableTransferAlphaValues == transferfunctionCheckBox1.Checked) return;
            GameProperties.Instance.enableTransferAlphaValues = transferfunctionCheckBox1.Checked;
            updateXnaEngineState(false, false, true, true);
            #endregion
          
        }

        /// <summary>
        /// Handle to enable the Show Histogram checkbox.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleTransferFunctionCheckbox2(object sender, EventArgs e)
        {
            GameProperties.Instance.showDensityValue = transferfunctionCheckBox2.Checked;
            TransferFunctionPanel.DensityFunction = transferfunctionCheckBox2.Checked;
            TransferFunctionPanel.Refresh();
        }


        /// <summary>
        /// Handle for the mouse down event in the transferfunction panel.
        /// If MouseButtons.Left -> Changes control point position, if any;
        /// If MouseButtons.Right -> Adds new control point, or change current one, opens a color piker.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void TransferFunctionPanel_MouseDown(object sender, MouseEventArgs e)
        {
            if (!transferfunctionCheckBox.Checked) return;

            Point newPos = clampToPanelDimensions(e.X, e.Y);

            if (e.Button == MouseButtons.Left) //select control point to move
            {
                tfPanelMouseDown = true;
                tfPanelMouseDownPos = newPos;
            }

            #region create new control point or edit one
            if (e.Button == MouseButtons.Right) //create new control point or edit one
            {
                ColorDialog colorPicker = new ColorDialog();
                //if user pick new color
                if (colorPicker.ShowDialog() == DialogResult.OK)
                {
                    TransferFunctionPanel.setControlPoint(newPos, colorPicker.Color);
                    TransferFunctionPanel.SetControlPoints = true;
                    TransferFunctionPanel.Refresh();
                    updateXnaEngineState(false, false, true, true); 
                }//end if ok
            } //end if right button
            #endregion

        }

        /// <summary>
        /// Handle for the mouse move event in the transferfunction panel.
        /// Changes position(-> iso value and alpha) of the selected control point, if any.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void TransferFunctionPanel_MouseMove(object sender, MouseEventArgs e)
        {
            if (!transferfunctionCheckBox.Checked) return;

            //Console.WriteLine("e.x " + e.X);
            //Console.WriteLine("e.y " + e.Y);
            if (tfPanelMouseDown)
            {

                Point newPos = clampToPanelDimensions(e.X, e.Y);
                if (TransferFunctionPanel.changeControlPoint(tfPanelMouseDownPos, newPos))
                {
                    TransferFunctionPanel.SetControlPoints = true;
                    TransferFunctionPanel.Refresh();
                    //update start pos for this control point
                    tfPanelMouseDownPos = newPos;
                }
              
            }//end if

        }

        /// <summary>
        /// Ends the add/edit control point interaction for the tf control
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void TransferFunctionPanel_MouseUp(object sender, MouseEventArgs e)
        {

            tfPanelMouseDown = false;
            updateXnaEngineState(false, false, true, true);
        }

        /// <summary>
        /// Removes a selected control point in the cp control.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void TransferFunctionPanel_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            if (!transferfunctionCheckBox.Checked) return;

            Point cpPos = clampToPanelDimensions(e.X, e.Y);
            if(TransferFunctionPanel.removeControlPoint(cpPos))
            {
                TransferFunctionPanel.SetControlPoints = true;
                TransferFunctionPanel.Refresh();
                 updateXnaEngineState(false, false, true, true);
            }

        }
        
        #endregion

        #region extended features
        /// <summary>
        /// Enables/Disables the cvd functionality
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleEnableCVDCheckBox(object sender, EventArgs e)
        {
            if (enableCVDCheckBox.Checked)
            {
                this.cvdViewCheckBox1.Checked = GameProperties.Instance.originalView;
                this.cvdViewCheckBox2.Checked = GameProperties.Instance.recoloredView;
                this.cvdViewCheckBox3.Checked = GameProperties.Instance.cvdView;
                this.cvdViewCheckBox4.Checked = GameProperties.Instance.simulate;
                this.cvdViewCheckBox1.Enabled = true;
                this.cvdViewCheckBox2.Enabled = true;
                this.cvdViewCheckBox3.Enabled = true;
                this.cvdViewCheckBox4.Enabled = true;
                this.cvdRadioButton1.Enabled = true;
                this.cvdRadioButton2.Enabled = true;
                this.cvdRadioButton3.Enabled = true;
                //this.cvdViewToolStripMenuItem.Enabled = true;
                //this.recoloredViewToolStripMenuItem.Enabled = true;
            }
            else
            {
                this.cvdViewCheckBox1.Enabled = false;
                this.cvdViewCheckBox2.Enabled = false;
                this.cvdViewCheckBox3.Enabled = false;
                this.cvdViewCheckBox4.Enabled = false;
                this.cvdRadioButton1.Enabled = false;
                this.cvdRadioButton2.Enabled = false;
                this.cvdRadioButton3.Enabled = false;
                //this.cvdViewToolStripMenuItem.Enabled = false;
                //this.recoloredViewToolStripMenuItem.Enabled = false;
            }
            GameProperties.Instance.enableCVDView = enableCVDCheckBox.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Enables/Disables the cvd view checkboxes;
        /// Sets the original rgb image view == 3d view mode
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleEnableCVDViewCheckBox1(object sender, EventArgs e)
        {
            if (this.cvdViewCheckBox1.Checked)
            {
                volViewToolStripMenuItem.Checked = true;
                GameProperties.Instance.viewMode = ViewMode.VolView;
            }
            else
            {  
                volViewToolStripMenuItem.Checked = false;
            }

            GameProperties.Instance.originalView = cvdViewCheckBox1.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Enables/Disables the cvd view checkboxes;
        /// Sets the recolored image view.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleEnableCVDViewCheckBox2(object sender, EventArgs e)
        {
            if (this.cvdViewCheckBox2.Checked)
            {
                
                this.recoloredViewToolStripMenuItem.Checked = true;
                GameProperties.Instance.triggerRecolor = true;
            }
            else
            {
                this.recoloredViewToolStripMenuItem.Checked = false;
                GameProperties.Instance.triggerRecolor = false;
            }

            GameProperties.Instance.recoloredView = cvdViewCheckBox2.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Enables/Disables the cvd view checkboxes;
        /// Sets the cvd friendly view mode.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleEnableCVDViewCheckBox3(object sender, EventArgs e)
        {
            if (this.cvdViewCheckBox3.Checked)
            {
               //this.cvdViewToolStripMenuItem.Checked = true;
               GameProperties.Instance.triggerOptimize = true;
               GameProperties.Instance.CVDTFOptimized = false;
               GameProperties.Instance.cvdView = true;
            }
            else
            {
                //this.cvdViewToolStripMenuItem.Checked = false;
                GameProperties.Instance.triggerOptimize = false;
                GameProperties.Instance.cvdView = false;
            }

            //GameProperties.Instance.cvdView = cvdViewCheckBox3.Checked;
            updateXnaEngineState(true, false, true, true);
        }


        /// <summary>
        /// Enables/Disables the simulated image checkboxes;
        /// Sets the cvd friendly view mode.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleEnableCVDViewCheckBox4(object sender, EventArgs e)
        {
            if (this.cvdViewCheckBox4.Checked == true)
            {
               //trigger simulate
                GameProperties.Instance.triggerSimulate = true;
            }
            else
            {
                GameProperties.Instance.triggerSimulate = false;
            }

            GameProperties.Instance.simulatedView = cvdViewCheckBox4.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Enables/Disables the maximum intensity projection
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleMaxIntensityProjCheckBox(object sender, EventArgs e)
        {
            if (MaxIntensityProjCheckBox.Checked)
            {
                maximumIntensityProjectionToolStripMenuItem.Checked = true;
            }
            else
            {
                maximumIntensityProjectionToolStripMenuItem.Checked = false;
            }
            GameProperties.Instance.maximumIntensityProjection = MaxIntensityProjCheckBox.Checked;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Switsches between perpective and orthogonal projection.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handlePerspProjRadioButt(object sender, EventArgs e)
        {
            perspectiveProjectionToolStripMenuItem.Checked = true;
            orthogonalProjectionToolStripMenuItem.Checked = false;
            GameProperties.Instance.perspectiveProjection = true;
            updateXnaEngineState(true, false, true, true);
        }


        /// <summary>
        /// Switsches between the three cvd types: protanopy, deuteranopy, tritanopy.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleCVDTypeRadioButtons(object sender, EventArgs e)
        {
            /*if (sender == this.cvdRadioButton1)
            {
                this.cvdRadioButton2.Checked = !this.cvdRadioButton1.Checked;
                this.cvdRadioButton3.Checked = !this.cvdRadioButton1.Checked;
            }
            else if (sender == this.cvdRadioButton2)
            {
                this.cvdRadioButton1.Checked = !this.cvdRadioButton2.Checked;
                this.cvdRadioButton3.Checked = !this.cvdRadioButton3.Checked;
            }
            else if (sender == this.cvdRadioButton3)
            {
                this.cvdRadioButton1.Checked = !this.cvdRadioButton3.Checked;
                this.cvdRadioButton2.Checked = !this.cvdRadioButton3.Checked;
            }*/

            if (this.cvdRadioButton1.Checked)
            {
                GameProperties.Instance.cvdType = CVDType.Deuteranopy;
                GameProperties.Instance.CVD = "deuteranopy";
                GameProperties.Instance.triggerRecolor = true;
                GameProperties.Instance.triggerOptimize = true;
                GameProperties.Instance.CVDTFOptimized = false;
            }
            else if (this.cvdRadioButton2.Checked) 
            {
                GameProperties.Instance.cvdType = CVDType.Protanopy;
                GameProperties.Instance.CVD = "protanopy";
                GameProperties.Instance.triggerRecolor = true;
                GameProperties.Instance.triggerOptimize = true;
                GameProperties.Instance.CVDTFOptimized = false;
            }
            else if (this.cvdRadioButton3.Checked)
            {
                GameProperties.Instance.cvdType = CVDType.Tritanopy;
                GameProperties.Instance.CVD = "tritanopy";
                GameProperties.Instance.triggerRecolor = true;
                GameProperties.Instance.triggerOptimize = true;
                GameProperties.Instance.CVDTFOptimized = false;
            }

            updateXnaEngineState(true, false, true, true);
        }


        /// <summary>
        /// Switsches between perpective and orthogonal projection.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleOrthProjRadioButt(object sender, EventArgs e)
        {
            perspectiveProjectionToolStripMenuItem.Checked = false;
            orthogonalProjectionToolStripMenuItem.Checked = true;
            GameProperties.Instance.perspectiveProjection = false;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Enables Front to Back Blending
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleBlendRadioButt1(object sender, EventArgs e)
        {
            frontToBackBlendingToolStripMenuItem.Checked = true;
            backToFrontBlendingToolStripMenuItem.Checked = false;
            GameProperties.Instance.enableFrontToBackBlend = true;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        /// Enables Back To Front Blending
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleBlendRadioButt2(object sender, EventArgs e)
        {
            frontToBackBlendingToolStripMenuItem.Checked = false;
            backToFrontBlendingToolStripMenuItem.Checked = true;
            GameProperties.Instance.enableFrontToBackBlend = false;
            updateXnaEngineState(true, false, true, true);
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleShadingCheckBox(object sender, EventArgs e)
        {
            if (ShadingCheckBox.Checked)
            {
                shadingToolStripMenuItem.Checked = true;
            }
            else
            {
                shadingToolStripMenuItem.Checked = false;
            }
            GameProperties.Instance.enableShading = ShadingCheckBox.Checked;
            updateXnaEngineState(true, false, true, true); 
        }

        #endregion

        #endregion

        #endregion

        #region handle xna draw surface events
        /// <summary>
        /// Handle for the engine control. 
        /// Changes the view mode on double click.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleXNADrawSurface_MouseDoubleClick(object sender, MouseEventArgs e)
        {

            switch (GameProperties.Instance.viewMode)
            {
                case ViewMode.VolView:
                    GameProperties.Instance.viewMode = ViewMode.TopView;
                    volViewToolStripMenuItem.Checked = false;
                    topViewToolStripMenuItem.Checked = true;
                    sideViewToolStripMenuItem.Checked = false;
                    frontViewToolStripMenuItem.Checked = false;
                    break;
                case ViewMode.TopView:
                    GameProperties.Instance.viewMode = ViewMode.SideView;
                    volViewToolStripMenuItem.Checked = false;
                    topViewToolStripMenuItem.Checked = false;
                    sideViewToolStripMenuItem.Checked = true;
                    frontViewToolStripMenuItem.Checked = false;
                    break;
                case ViewMode.SideView:
                    GameProperties.Instance.viewMode = ViewMode.FrontView;
                    volViewToolStripMenuItem.Checked = false;
                    topViewToolStripMenuItem.Checked = false;
                    sideViewToolStripMenuItem.Checked = false;
                    frontViewToolStripMenuItem.Checked = true;
                    break;
                case ViewMode.FrontView:
                    GameProperties.Instance.viewMode = ViewMode.VolView;
                    volViewToolStripMenuItem.Checked = true;
                    topViewToolStripMenuItem.Checked = false;
                    sideViewToolStripMenuItem.Checked = false;
                    frontViewToolStripMenuItem.Checked = false;
                    break;
                default:
                    GameProperties.Instance.viewMode = ViewMode.VolView;
                    volViewToolStripMenuItem.Checked = true;
                    topViewToolStripMenuItem.Checked = false;
                    sideViewToolStripMenuItem.Checked = false;
                    frontViewToolStripMenuItem.Checked = false;
                    break;
            }

           updateXnaEngineState(true, false, true, true);
            
        }

        
        /// <summary>
        /// Handle for the engine control.
        /// Registers the mouse down event on the xna draw surface area 
        /// and holds the mouse position.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleXNADrawSurface_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left || e.Button == MouseButtons.Right) 
            {
                XnaDrawSurface_MouseDownEvent = true;
                mouseDownPosX = (int)MathHelper.Clamp(e.X, 0, XNADrawSurface.Width-1);
                mouseDownPosY = (int)MathHelper.Clamp(e.Y, 0, XNADrawSurface.Height - 1);
            }

        }

        /// <summary>
        /// Handle for the engine control.
        /// Registers the mouse down up on the xna draw surface area
        /// and resets the mouse position.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleXNADrawSurface_MouseUp(object sender, MouseEventArgs e)
        {
            mouseDownPosX = 0;
            mouseDownPosY = 0;
            mouseDownPosZ = 0;
            GameProperties.Instance.cameraInput = Vector3.Zero;
            XnaDrawSurface_MouseDownEvent = false;
        }

        /// <summary>
        /// Handle for the engine control.
        /// Navigates the camera (3d view mode).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleXNADrawSurface_MouseMove(object sender, MouseEventArgs e)
        {
            if (XnaDrawSurface_MouseDownEvent)
            {
                int nextPosX = (int)MathHelper.Clamp(e.X, 0, XNADrawSurface.Width - 1);
                int nextPosY = (int)MathHelper.Clamp(e.Y, 0, XNADrawSurface.Height - 1);

                int moveX = nextPosX - mouseDownPosX; 
                int moveY = mouseDownPosY - nextPosY; 

                //Console.WriteLine("Move X:" + moveX + "; Move Y:" + moveY);
                if (e.Button == MouseButtons.Left)
                {
                    GameProperties.Instance.cameraInput.X = moveX;
                    GameProperties.Instance.cameraInput.Y = moveY;
                }
                else if (e.Button == MouseButtons.Right)
                {
                    GameProperties.Instance.cameraInput.Z = moveY;
                }
                mouseDownPosX = nextPosX;
                mouseDownPosY = nextPosY;
                updateXnaEngineState(true, false, true, true);
            }
        }

        /// <summary>
        /// Handle for the engine control.
        /// Zoom in/out the camera.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleXNADrawSurface_MouseScroll(object sender, MouseEventArgs e)
        {
            if (XnaDrawSurface_MouseDownEvent)
            {
                //Console.WriteLine("Delta: " + e.Delta);
                //Console.WriteLine("System info scroll data: " + SystemInformation.MouseWheelScrollDelta);
                int numberLinesToMove = e.Delta * SystemInformation.MouseWheelScrollLines / 120;
                GameProperties.Instance.cameraInput.Z = numberLinesToMove;
                Console.WriteLine("Camera zoom: " + GameProperties.Instance.cameraInput.Z);
                updateXnaEngineState(true, false, true, true);
            }
        }

        /// <summary>
        /// Handle for the engine control.
        /// Moves the camera left/right - up/down.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void XNADrawSurface_KeyPress(object sender, KeyPressEventArgs e)
        {
            //TODO
            /*switch (e.KeyCode)
            {
              // http://msdn.microsoft.com/en-us/library/system.windows.forms.control.keypress(v=vs.110).aspx 
            }*/
        }

        /// <summary>
        /// updates the current mouse position in the game properties to zoom +/- the camera,
        /// but only for the 3d viewport
        /// </summary>
        /// <param name="scrollStep">how many steps the mouse was scrolled + / -</param>
        public void updateMouseScroll(int scrollStep)
        {
            //int mousePos = getMousePos(mouseDownPosX, mouseDownPosY);

            //zoom +/-
            if (GameProperties.Instance.viewMode == ViewMode.VolView)
            {
                // : logic in the camera class
                //
                //scrollStep --> e.Delta, 120->max zoom value ex. 10
                //if endvalue > 0 -> zoom + , < 0 zoom -, = 0 nothing
                GameProperties.Instance.position.Z = scrollStep;
                // mouseDownPosZ = scrollStep;
            }
            //move slice
            else
            {
                if (GameProperties.Instance.viewMode == ViewMode.TopView)
                {
                    //GameProperties.Instance.ySliceValue = -scrollStep/10;
                }
                else if (GameProperties.Instance.viewMode == ViewMode.SideView)
                {
                    //GameProperties.Instance.xSliceValue = -scrollStep/10;
                }
                else if (GameProperties.Instance.viewMode == ViewMode.FrontView)
                {
                    //GameProperties.Instance.zSliceValue = -scrollStep/10;
                }

            }

        }
        #endregion

        #region add functions

        /// <summary>
        /// Updates the transfer function states and call refresh for the tf control if we need to update or redraw.
        /// </summary>
        /// <param name="enableTF">Enables/disables the transferfunction.</param>
        /// <param name="drawDF">Draws the density function of the currrent dataset.</param>
        /// <param name="setCP">True when new control points was added in the tf control.</param>
        /// <param name="getCP">True when new dataset was loaded -> set initial control points; 
        /// or to read a control points definition from a xml file.</param>
        /// <param name="updateDF">True when new dataset was loaded -> update the density values.</param>
        public void updateTFState(bool enableTF, bool drawDF, bool setCP, bool getCP, bool updateDF)
        {
            this.TransferFunctionPanel.EnableTransferFunction = enableTF;
            this.TransferFunctionPanel.DensityFunction = drawDF;
            if (setCP && getCP) { setCP = false; getCP = true; } //if both true -> error -> load initial control points
            this.TransferFunctionPanel.SetControlPoints = setCP;
            this.TransferFunctionPanel.GetControlPoints = getCP;
            this.TransferFunctionPanel.UpdateIsoValues = updateDF;
            if (enableTF || drawDF || setCP || getCP || updateDF) this.TransferFunctionPanel.Refresh();
        }
        
        /// <summary>
        /// Updates the engine states and call refresh for the engine control if we need to redraw.
        /// </summary>
        /// <param name="updateCam"></param>
        /// <param name="load"></param>
        /// <param name="update"></param>
        /// <param name="draw"></param>
        public void updateXnaEngineState(bool updateCam, bool load, bool update, bool draw)
        {
            GameProperties.Instance.engineState.updateCamera = updateCam;
            GameProperties.Instance.engineState.loadData = load;
            GameProperties.Instance.engineState.updateData = update;
            GameProperties.Instance.engineState.drawData = draw;
            if (updateCam || load || update || draw) this.XNADrawSurface.Refresh();
        }

        /// <summary>
        /// Loads transferfunction definition from a xml file.
        /// </summary>
        /// <param name="fileName"></param>
        public void loadTransferFunction(String fileName)
        {
            GameProperties.Instance.colorControlPoints.Clear();
            GameProperties.Instance.alphaControlPoints.Clear();

            XmlTextReader xmlReader = new XmlTextReader(fileName);
            xmlReader.MoveToElement();

            float r = 0.0f;
            float g = 0.0f;
            float b = 0.0f;
            float a = 0.0f;
            int isoValue = 0;

            while (xmlReader.Read())
            {
                if (xmlReader.NodeType == XmlNodeType.Element && xmlReader.Name.Equals("ControlPoint"))
                {
                    while (!(xmlReader.NodeType == XmlNodeType.EndElement && xmlReader.Name.Equals("ControlPoint")))
                    {
                        xmlReader.Read();

                        if (xmlReader.NodeType.Equals(XmlNodeType.Element))
                        {
                            String elementName = xmlReader.Name;
                            xmlReader.Read();
                            if (xmlReader.NodeType.Equals(XmlNodeType.Text))
                            {
                                switch (elementName)
                                {
                                    case "ColorR":
                                        r = float.Parse(xmlReader.Value);
                                        break;

                                    case "ColorG":
                                        g = float.Parse(xmlReader.Value);
                                        break;

                                    case "ColorB":
                                        b = float.Parse(xmlReader.Value);
                                        break;

                                    case "ColorA":
                                        a = float.Parse(xmlReader.Value);
                                        break;

                                    case "IsoValue":
                                        isoValue = int.Parse(xmlReader.Value);
                                        break;
                                }//end switch
                            }//end if
                        }
                    }//end while

                    //save new control point
                    GameProperties.Instance.colorControlPoints.Add(new TransferControlPoint(r, g, b, isoValue));
                    GameProperties.Instance.alphaControlPoints.Add(new TransferControlPoint(a, isoValue));
                }
            }
        }

        /// <summary>
        /// saves all control points that are set 
        /// </summary>
        public void saveTransferFunction(string fileName)
        {
            XmlTextWriter xmlWriter = new XmlTextWriter(fileName, null);

            xmlWriter.WriteStartDocument();
            xmlWriter.WriteStartElement("TransferFunction");

            for (int i = 0; i < GameProperties.Instance.colorControlPoints.Count(); i++)
            {
                xmlWriter.WriteStartElement("ControlPoint");

                xmlWriter.WriteStartElement("ColorR");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].color.X.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("ColorG");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].color.Y.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("ColorB");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].color.Z.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("ColorA");
                xmlWriter.WriteString(GameProperties.Instance.alphaControlPoints[i].color.W.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("IsoValue");
                xmlWriter.WriteString(GameProperties.Instance.colorControlPoints[i].isoValue.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteEndElement();
            }

            xmlWriter.WriteEndElement();
            xmlWriter.WriteEndDocument();
            xmlWriter.Close();
        }

        private Point clampToPanelDimensions(int posX, int posY)
        {
            Point p = new Point(posX, posY);
            if (p.X < 0) { p.X = 0; }
            else if (p.X > this.TransferFunctionPanel.Width - 1)
            { p.X = this.TransferFunctionPanel.Width - 1; }
            if (p.Y < 0) { p.Y = 0; }
            else if (p.Y > this.TransferFunctionPanel.Height - 1)
            { p.Y = this.TransferFunctionPanel.Height - 1; }
            return p;
        }
        #endregion



        protected override void WndProc(ref Message m)
        {

            // WM_SYSCOMMAND
            if (m.Msg == 0x0112)
            {
                if (m.WParam == new IntPtr(0xF030) // Maximize event - SC_MAXIMIZE from Winuser.h
                    || m.WParam == new IntPtr(0xF120)) // Restore event - SC_RESTORE from Winuser.h
                {

                    updateXnaEngineState(true, false, true, true);
                    //Console.WriteLine("Size:" + Size);
                }
            }

            base.WndProc(ref m);
        }

        private void toolStripMenuItem3_Click(object sender, EventArgs e)
        {

        }

       
    }
 }
