﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace visLU.Gui
{
    /// <summary>
    /// Form for the slicing panel.
    /// </summary>
    public partial class SlicingForm : Form
    {

        private MainWindow mainForm;

        /// <summary>
        ///
        /// </summary>
        public SlicingForm()
        {
            InitializeComponent();
        }

        /// <summary>
        ///
        /// </summary>
        public MainWindow MainForm
        {
            set { mainForm = value; }
        }

        /// <summary>
        /// Initializes the slider in slicing form with the values from the main form.
        /// </summary>
        /// <param name="slice"></param>
        /// <param name="increment"></param>
        /// <param name="decimalPlaces"></param>
        /// <param name="value"></param>
        public void setSlicingGroups(int slice, Decimal increment, int decimalPlaces, Decimal value)
        {
            switch (slice)
            {
                case 1:
                    XSliceNumericUpDown.Increment = increment;
                    XSliceNumericUpDown.DecimalPlaces = decimalPlaces;
                    XSliceNumericUpDown.Value = value;
                    break;
                case 2:
                    YSliceNumericUpDown.Increment = increment;
                    YSliceNumericUpDown.DecimalPlaces = decimalPlaces;
                    YSliceNumericUpDown.Value = value;
                    break;
                case 3:
                    ZSliceNumericUpDown.Increment = increment;
                    ZSliceNumericUpDown.DecimalPlaces = decimalPlaces;
                    ZSliceNumericUpDown.Value = value;
                    break;
                default:
                    break;

            }

        }

        /// <summary>
        /// Initializes the slider in slicing form with the values from the main form.
        /// </summary>
        /// <param name="slice"></param>
        /// <param name="maximum"></param>
        /// <param name="value"></param>
        public void setSlicingGroups(int slice, int maximum, int value)
        {
            switch (slice)
            {
                case 1:
                    XSlider.Maximum = maximum;
                    XSlider.Value = value;
                    break;
                case 2:
                    YSlider.Maximum = maximum;
                    YSlider.Value = value;
                    break;
                case 3:
                    ZSlider.Maximum = maximum;
                    ZSlider.Value = value;
                    break;
                default:
                    break;

            }

        }

        #region event handler
        /// <summary>
        ///
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleSliceNumericUpDown(object sender, EventArgs e)
        {
            if (sender == XSliceNumericUpDown)
            {
                XSlider.Value = Decimal.ToInt32(Decimal.Multiply(XSliceNumericUpDown.Value, new Decimal(XSlider.Maximum)));
                GameProperties.Instance.xSliceValue = (float)XSliceNumericUpDown.Value;

            }
            else if (sender == YSliceNumericUpDown)
            {
                YSlider.Value = Decimal.ToInt32(Decimal.Multiply(YSliceNumericUpDown.Value, new Decimal(YSlider.Maximum)));
                GameProperties.Instance.ySliceValue = (float)YSliceNumericUpDown.Value;

            }
            else if (sender == ZSliceNumericUpDown)
            {
                ZSlider.Value = Decimal.ToInt32(Decimal.Multiply(ZSliceNumericUpDown.Value, new Decimal(ZSlider.Maximum)));
                GameProperties.Instance.zSliceValue = (float)ZSliceNumericUpDown.Value;

            }
            else { return; }

            mainForm.setSlicingGroups(1, XSliceNumericUpDown.Increment, XSliceNumericUpDown.DecimalPlaces, XSliceNumericUpDown.Value);
            mainForm.setSlicingGroups(2, YSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, YSliceNumericUpDown.Value);
            mainForm.setSlicingGroups(3, ZSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, ZSliceNumericUpDown.Value);
            mainForm.setSlicingGroups(1, XSlider.Maximum, XSlider.Value);
            mainForm.setSlicingGroups(2, YSlider.Maximum, YSlider.Value);
            mainForm.setSlicingGroups(3, ZSlider.Maximum, ZSlider.Value);
            mainForm.updateSliceGroups();

            if (!GameProperties.Instance.viewMode.Equals(ViewMode.VolView))
            {
                mainForm.updateXnaEngineState(false, false, false, true);
            }
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleScrollSlider(object sender, EventArgs e)
        {
            if (sender == XSlider)
            {
                Decimal newValue = Decimal.Divide(new Decimal(XSlider.Value), new Decimal(XSlider.Maximum));
                XSliceNumericUpDown.Value = newValue;
                GameProperties.Instance.xSliceValue = (float)newValue;

            }
            else if (sender == YSlider)
            {
                Decimal newValue = Decimal.Divide(new Decimal(YSlider.Value), new Decimal(YSlider.Maximum));
                YSliceNumericUpDown.Value = newValue;
                GameProperties.Instance.ySliceValue = (float)newValue;

            }
            else if (sender == ZSlider)
            {
                Decimal newValue = Decimal.Divide(new Decimal(ZSlider.Value), new Decimal(ZSlider.Maximum));
                ZSliceNumericUpDown.Value = newValue;
                GameProperties.Instance.zSliceValue = (float)newValue;

            }
            else { return; }

            mainForm.setSlicingGroups(1, XSliceNumericUpDown.Increment, XSliceNumericUpDown.DecimalPlaces, XSliceNumericUpDown.Value);
            mainForm.setSlicingGroups(2, YSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, YSliceNumericUpDown.Value);
            mainForm.setSlicingGroups(3, ZSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, ZSliceNumericUpDown.Value);
            mainForm.setSlicingGroups(1, XSlider.Maximum, XSlider.Value);
            mainForm.setSlicingGroups(2, YSlider.Maximum, YSlider.Value);
            mainForm.setSlicingGroups(3, ZSlider.Maximum, ZSlider.Value);
            mainForm.updateSliceGroups();

            if (!GameProperties.Instance.viewMode.Equals(ViewMode.VolView))
            {
                mainForm.updateXnaEngineState(false, false, false, true);
            }
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void handleSlicingFormOnClosing(object sender, FormClosingEventArgs e)
        {
                mainForm.setSlicingGroups(1, XSliceNumericUpDown.Increment, XSliceNumericUpDown.DecimalPlaces, XSliceNumericUpDown.Value);
                mainForm.setSlicingGroups(2, YSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, YSliceNumericUpDown.Value);
                mainForm.setSlicingGroups(3, ZSliceNumericUpDown.Increment, YSliceNumericUpDown.DecimalPlaces, ZSliceNumericUpDown.Value);
                mainForm.setSlicingGroups(1, XSlider.Maximum, XSlider.Value);
                mainForm.setSlicingGroups(2, YSlider.Maximum, YSlider.Value);
                mainForm.setSlicingGroups(3, ZSlider.Maximum, ZSlider.Value);
                mainForm.updateSliceGroups();
                mainForm.SliceFormOpen = false;
        }

        #endregion
    }
}
